import pytest
from tests.metaschema.datatypes.test_MetaschemaType import (
    TestMetaschemaType as base_class)
from yggdrasil.metaschema.datatypes.FunctionMetaschemaType import example_func


class TestFunctionMetaschemaType(base_class):
    r"""Test class for FunctionMetaschemaType class with float."""

    _mod = 'yggdrasil.metaschema.datatypes.FunctionMetaschemaType'
    _cls = 'FunctionMetaschemaType'

    @pytest.fixture(scope="class")
    def value(self):
        r"""function: Test function."""
        return example_func
    
    @pytest.fixture(scope="class")
    def valid_encoded(self, python_class, typedef_base):
        r"""list: Encoded objects that are valid under this type."""
        return [dict(typedef_base,
                     type=python_class.name)]
    
    @pytest.fixture(scope="class")
    def valid_decoded(self, value):
        r"""list: Objects that are valid under this type."""
        return [value]
    
    @pytest.fixture(scope="class")
    def invalid_decoded(self):
        r"""list: Objects that are invalid under this type."""
        return [object]

    @pytest.fixture(scope="class")
    def compatible_objects(self, value):
        r"""list: Objects that are compatible with this type."""
        return [(value, value, None)]

    @pytest.fixture(scope="class")
    def valid_normalize(self):
        r"""list: Pairs of pre-/post-normalized objects."""
        return [(None, None),
                ('yggdrasil.metaschema.datatypes.FunctionMetaschemaType'
                 ':example_func', example_func)]

    def test_decode_data_errors(self, python_class):
        r"""Test errors in decode_data."""
        with pytest.raises(ValueError):
            python_class.decode_data('hello', None)
        with pytest.raises(AttributeError):
            python_class.decode_data('yggdrasil:invalid', None)



import pytest
from yggdrasil.communication import new_comm
from tests.communication.test_CommBase import BaseComm as base_class


class TestValueComm(base_class):
    r"""Test for ValueComm communication class."""

    comm = 'ValueComm'

    @pytest.fixture(scope="class", autouse=True)
    def commtype(self):
        r"""Communicator type being tested."""
        return "value"

    def create_recv_comm(self, name, commtype, use_async, testing_options,
                         **kwargs):
        r"""Create a recv comm."""
        kws = dict(commtype=commtype, reverse_names=True,
                   direction='recv', use_async=use_async)
        kws.update(testing_options['kwargs'])
        kws.update(kwargs)
        x = new_comm(name, **kws)
        assert(x.is_open)
        return x

    @pytest.fixture(scope="class")
    def global_send_comm(self):
        r"""Communicator for sending messages."""
        return None

    @pytest.fixture(scope="class")
    def global_recv_comm(self):
        r"""Communicator for receiving messages."""
        return None

    @pytest.fixture
    def send_comm(self):
        r"""Communicator for sending messages."""
        return None

    @pytest.fixture
    def recv_comm(self, name, commtype, use_async, testing_options,
                  sleep_after_connect, close_comm):
        r"""Communicator for receiving messages."""
        recv_comm = self.create_recv_comm(name, commtype, use_async,
                                          testing_options)
        if sleep_after_connect:
            recv_comm.sleep()
        yield recv_comm
        close_comm(recv_comm)

    @pytest.fixture(scope="class")
    def maxMsgSize(self):
        r"""int: Maximum message size."""
        return 0

    def test_send_recv(self, recv_comm, testing_options, use_async, timeout,
                       map_sent2recv, wait_on_function):
        r"""Test send/recv of a small message."""
        n_recv = testing_options['kwargs']['count']
        msg_recv = testing_options['msg']
        if use_async:
            wait_on_function(lambda: recv_comm.n_msg_recv > 0)
            assert(recv_comm.n_msg_recv > 0)
        else:
            assert(recv_comm.n_msg_recv == n_recv)
        # Wait for messages to be received
        for i in range(n_recv):
            flag, msg_recv0 = recv_comm.recv(timeout)
            assert(flag)
            assert(msg_recv0 == map_sent2recv(msg_recv))
        # Receive after empty
        assert(recv_comm.is_open)
        flag, msg_recv0 = recv_comm.recv(timeout)
        assert(not flag)
        assert(recv_comm.is_eof(msg_recv0))
        # assert(flag)
        # assert(recv_comm.is_empty_recv(msg_recv0))
        # Confirm recept of messages
        recv_comm.wait_for_confirm(timeout=timeout)
        assert(recv_comm.is_confirmed)
        recv_comm.confirm(noblock=True)
        assert(recv_comm.n_msg_recv == 0)

    def test_send_recv_after_close(self, recv_comm, use_async):
        r"""Test that opening twice dosn't cause errors and that send/recv after
        close returns false."""
        recv_comm.open()
        recv_comm.close()
        assert(recv_comm.is_closed)
        flag, msg_recv = recv_comm.recv()
        if not use_async:
            assert(not flag)
        with pytest.raises(RuntimeError):
            recv_comm.send(None)

    def test_purge(self, recv_comm, wait_on_function):
        r"""Test purging messages from the comm."""
        wait_on_function(lambda: recv_comm.n_msg > 0)
        assert(recv_comm.n_msg > 0)
        recv_comm.purge()
        assert(recv_comm.n_msg == 0)
        recv_comm.close()
        recv_comm.purge()



import pytest
from tests.communication.test_FileComm import TestFileComm as base_class


class TestAsciiFileComm(base_class):
    r"""Test for AsciiFileComm communication class."""

    @pytest.fixture(scope="class", autouse=True)
    def filetype(self):
        r"""Communicator type being tested."""
        return "ascii"

    def test_send_recv_comment(self, send_comm, recv_comm, testing_options):
        r"""Test send/recv with commented message."""
        msg_send = send_comm.serializer.comment + testing_options['msg']
        flag = send_comm.send(msg_send)
        assert(flag)
        flag, msg_recv = recv_comm.recv()
        assert(not flag)
        assert(msg_recv == recv_comm.eof_msg)



import os


def get_testing_options():
    r"""Get testing parameters for this example."""
    return dict(
        expected_output_files=[
            os.path.join('Output', 'outputB.txt'),
            os.path.join('Output', 'outputC.txt')],
        output_dir=None,
        output_files=['outputB.txt', 'outputC.txt'])



import pytest
from tests.metaschema.datatypes.test_JSONObjectMetaschemaType import (
    TestJSONObjectMetaschemaType as base_class)
from tests import TestClassBase as base_class_dict
import os
import copy
import shutil
import tempfile
import numpy as np
from yggdrasil.metaschema.datatypes import PlyMetaschemaType


def test_create_schema():
    r"""Test create_schema."""
    with pytest.raises(RuntimeError):
        PlyMetaschemaType.create_schema(overwrite=False)
    temp = os.path.join(tempfile.gettempdir(), 'temp_schema')
    old_schema = PlyMetaschemaType.get_schema()
    try:
        shutil.move(PlyMetaschemaType._schema_file, temp)
        new_schema = PlyMetaschemaType.get_schema()
        assert(old_schema == new_schema)
    except BaseException:  # pragma: debug
        shutil.move(temp, PlyMetaschemaType._schema_file)
        raise
    shutil.move(temp, PlyMetaschemaType._schema_file)


def test_translate_ply2fmt_errors():
    r"""Test errors in translate_ply2fmt."""
    with pytest.raises(ValueError):
        PlyMetaschemaType.translate_ply2fmt('invalid')


def test_translate_ply2py_errors():
    r"""Test errors in translate_ply2py."""
    with pytest.raises(ValueError):
        PlyMetaschemaType.translate_ply2py('invalid')


def test_translate_py2ply_errors():
    r"""Test errors in translate_py2ply."""
    with pytest.raises(ValueError):
        PlyMetaschemaType.translate_py2ply('float128')


def test_singular2plural():
    r"""Test conversion from singular element names to plural ones and back."""
    pairs = [('face', 'faces'), ('vertex', 'vertices'),
             ('vertex_index', 'vertex_indices')]
    for s, p in pairs:
        assert(PlyMetaschemaType.singular2plural(s) == p)
        assert(PlyMetaschemaType.plural2singular(p) == s)
    with pytest.raises(ValueError):
        PlyMetaschemaType.plural2singular('invalid')


class TestPlyDict(base_class_dict):
    r"""Test for PlyDict class."""
    
    _mod = 'yggdrasil.metaschema.datatypes.PlyMetaschemaType'
    _cls = 'PlyDict'

    @pytest.fixture(scope="class")
    def simple_test(self, ply_test_value_simple):
        r"""dict: Simple test structure."""
        return ply_test_value_simple

    @pytest.fixture
    def instance_kwargs(self, ply_test_value):
        r"""Keyword arguments for a new instance of the tested class."""
        return ply_test_value

    @pytest.fixture
    def objects_array_dict(self, instance):
        r"""Objects for testing transformation to/from dict of arrays."""
        orig = copy.deepcopy(instance)
        for f in orig['faces']:
            for k in ['red', 'green', 'blue']:
                f[k] = int(255)
        orig_no_color = copy.deepcopy(instance)
        for v in orig_no_color['vertices']:
            for k in ['red', 'green', 'blue']:
                v.pop(k, None)
        for e in orig_no_color['edges']:
            for k in ['red', 'green', 'blue']:
                e.pop(k, None)
        return [orig, orig_no_color]

    @pytest.fixture
    def objects_trimesh(self, instance):
        r"""Objects for testing transformation to/from trimesh class."""
        if not PlyMetaschemaType.trimesh:
            pytest.skip("trimesh not available")
        orig = copy.deepcopy(instance)
        for k in ['material', 'edges']:
            orig.pop(k, None)
        return [orig]

    @pytest.fixture(scope="class")
    def remove_hd_faces(self):
        r"""Remove higher dimension faces (>3 verts)."""
        def remove_hd_faces_w(obj):
            for f in obj.get('faces', []):
                f['vertex_index'] = f['vertex_index'][:3]
            return obj
        return remove_hd_faces_w

    def test_count_elements(self, instance):
        r"""Test count_elements."""
        with pytest.raises(ValueError):
            instance.count_elements('invalid')
        x = instance.count_elements('vertices')
        y = instance.count_elements('vertex')
        assert(x == y)

    def test_mesh(self, instance):
        r"""Test mesh."""
        instance.mesh

    def test_merge(self, instance):
        r"""Test merging two ply objects."""
        ply1 = copy.deepcopy(instance)
        ply2 = ply1.merge(instance)
        ply1.merge([instance], no_copy=True)
        assert(ply1 == ply2)

    def test_append(self, python_class, instance):
        r"""Test appending ply objects."""
        basic = python_class(vertices=instance['vertices'], faces=[])
        basic.append(instance)

    def test_apply_scalar_map(self, class_name, instance):
        r"""Test applying a scalar colormap."""
        o = copy.deepcopy(instance)
        scalar_arr = np.arange(o.count_elements('faces')).astype('float')
        with pytest.raises(NotImplementedError):
            o.apply_scalar_map(scalar_arr, scale_by_area=True)
        new_faces = []
        if 'Obj' in class_name:
            for f in o['faces']:
                if len(f) == 3:
                    new_faces.append(f)
        else:
            for f in o['faces']:
                if len(f['vertex_index']) == 3:
                    new_faces.append(f)
        o['faces'] = new_faces
        for scale in ['linear', 'log']:
            o2 = copy.deepcopy(o)
            o1 = o.apply_scalar_map(scalar_arr, scaling=scale, scale_by_area=True)
            o2.apply_scalar_map(scalar_arr, scaling=scale, scale_by_area=True,
                                no_copy=True)
            assert(o1 == o2)

    @pytest.fixture
    def to_from_scene(self, python_class, instance):
        r"""Perform convertions to/from PlantGL scene."""
        from yggdrasil.drivers.LPyModelDriver import LPyModelDriver
        if not LPyModelDriver.is_installed():
            pytest.skip("LPy library not installed.")
            
        def to_from_scene_w(data):  # pragma: lpy
            o1 = instance
            cls = o1.__class__
            s = o1.to_scene(name='test')
            o2 = cls.from_scene(s)
            # Direct equivalence won't happen unless test is just for simple
            # mesh as faces with more than 3 vertices will be triangulated.
            cls = python_class
            o1 = cls(data)
            s = o1.to_scene(name='test')
            o2 = cls.from_scene(s)
            # import pprint
            # print('o2')
            # pprint.pprint(o2)
            # print('o1')
            # pprint.pprint(o1)
            assert(o2 == o1)
        return to_from_scene_w

    def test_to_from_scene(self, to_from_scene, simple_test):  # pragma: lpy
        r"""Test conversion to/from PlantGL scene."""
        to_from_scene(simple_test)

    def test_to_from_dict(self, python_class, instance):
        r"""Test transformation to/from dict."""
        x = instance.as_dict()
        y = python_class.from_dict(x)
        assert(y == instance)

    def test_to_from_array_dict(self, python_class, objects_array_dict):
        r"""Test transformation to/from dict of arrays."""
        for y0 in objects_array_dict:
            x = y0.as_array_dict()
            y = python_class.from_array_dict(x)
            assert(y == y0)

    def test_to_from_trimesh(self, python_class, objects_trimesh,
                             remove_hd_faces):
        r"""Test transformation to/from trimesh class."""
        if not PlyMetaschemaType.trimesh:
            pytest.skip("trimesh not available")
        for y0 in objects_trimesh:
            y0 = remove_hd_faces(y0)
            x = y0.as_trimesh()
            y = python_class.from_trimesh(x)
            assert(y == y0)

    def test_properties(self, instance):
        r"""Test explicit exposure of specific element counts as properties
        against counts based on singular elements."""
        instance.bounds
        assert(instance.nvert == instance.count_elements('vertex'))
        assert(instance.nface == instance.count_elements('face'))


class TestPlyMetaschemaType(base_class):
    r"""Test class for PlyMetaschemaType class."""

    _mod = 'yggdrasil.metaschema.datatypes.PlyMetaschemaType'
    _cls = 'PlyMetaschemaType'

    @pytest.fixture(scope="class")
    def value(self, ply_test_value):
        r"""list: Test value."""
        return ply_test_value

    @pytest.fixture(scope="class")
    def fulldef(self, python_class):
        r"""dict: Full type definitions."""
        return {'type': python_class.name}
    
    @pytest.fixture(scope="class")
    def typedef_base(self, python_class):
        r"""dict: Base type definition."""
        return {'type': python_class.name}

    @pytest.fixture(scope="class")
    def valid_encoded(self, fulldef):
        r"""list: Encoded objects that are valid under this type."""
        return [fulldef]
        
    @pytest.fixture(scope="class")
    def valid_decoded(self, value, ply_test_value, ply_test_value_int64):
        r"""list: Objects that are valid under this type."""
        out = [value,
               PlyMetaschemaType.PlyDict(**ply_test_value),
               {'vertices': [], 'faces': [],
                'alt_verts': copy.deepcopy(ply_test_value['vertices'])},
               ply_test_value_int64]
        if PlyMetaschemaType.trimesh:
            out.append(PlyMetaschemaType.PlyDict(**ply_test_value).as_trimesh())
        return out
            
    @pytest.fixture(scope="class")
    def invalid_encoded(self):
        r"""list: Encoded objects that are invalid under this type."""
        return [{}]

    @pytest.fixture(scope="class")
    def invalid_decoded(self):
        r"""list: Objects that are invalid under this type."""
        return [{'vertices': [{k: 0.0 for k in 'xyz'}],
                 'faces': [{'vertex_index': [0, 1, 2]}]}]

    @pytest.fixture(scope="class")
    def compatible_objects(self, value):
        r"""list: Objects that are compatible with this type."""
        return [(value, value, None)]

    @pytest.fixture(scope="class")
    def encode_data_kwargs(self):
        r"""dict: Keyword arguments for encoding data of this type."""
        return {'comments': ['Test comment']}

    @pytest.fixture
    def nested_result(self, nested_approx):
        r"""Prepare value for comparison."""
        def nested_result_w(x):
            if PlyMetaschemaType.trimesh:
                if isinstance(x, PlyMetaschemaType.trimesh.base.Trimesh):
                    x = PlyMetaschemaType.PlyDict.from_trimesh(x)
            return nested_approx(x)
        return nested_result_w
    
    def test_decode_data_errors(self, python_class):
        r"""Test errors in decode_data."""
        with pytest.raises(ValueError):
            python_class.decode_data('hello', None)



from yggdrasil.communication import BufferComm


def test_LockedBuffer_clear(wait_on_function):
    r"""Test the clear method of the LockedBuffer class."""
    x = BufferComm.LockedBuffer()
    x.append('test')
    wait_on_function(lambda: not x.empty())
    x.clear()
    assert(len(x) == 0)


def test_LockedBuffer_pop():
    r"""Test the pop method of the LockedBuffer class."""
    x = BufferComm.LockedBuffer()
    assert(x.pop(default='hello') == 'hello')



import pytest
import os
import logging
from yggdrasil import runner
from yggdrasil.drivers import MatlabModelDriver
from yggdrasil.examples import yamls as ex_yamls
from tests.drivers.test_ModelDriver import TestModelDriver as base_class


logger = logging.getLogger(__name__)


@pytest.fixture(scope="session")
def matlab_session_file():
    return os.path.join(os.getcwd(), 'nt_screen_session.txt')


@pytest.mark.related_language('matlab')
def test_is_matlab_running():
    r"""Test if there is Matlab engine running."""
    MatlabModelDriver.is_matlab_running()
    MatlabModelDriver.kill_all()
    assert(not MatlabModelDriver.is_matlab_running())


@pytest.mark.skipif(MatlabModelDriver._matlab_engine_installed,
                    reason="Matlab installed.")
def test_matlab_engine_not_installed():  # pragma: no matlab
    r"""Assert that errors are raised when Matlab engine is not installed."""
    with pytest.raises(RuntimeError):
        MatlabModelDriver.start_matlab_engine()
    with pytest.raises(RuntimeError):
        MatlabModelDriver.stop_matlab_engine(None, None, None, None)
    with pytest.raises(RuntimeError):
        MatlabModelDriver.MatlabProcess(None, None)


@pytest.mark.language('matlab')
def test_matlab_runner():  # pragma: matlab
    r"""Use get_runner to start a Matlab run."""
    cr = runner.get_runner([ex_yamls['hello']['matlab']])
    cr.run()


@pytest.mark.language('matlab')
def test_matlab_exit(scripts):  # pragma: matlab
    r"""Test error when model contains 'exit' call."""
    MatlabModelDriver.MatlabModelDriver('error', [scripts['matlab_error']])
    # Re-enable if it becomes necessary to raise an error instead of just a warning
    # with pytest.raises(RuntimeError):
    #     MatlabModelDriver.MatlabModelDriver('error', [scripts['matlab_error']])


@pytest.mark.language('matlab')
def test_locate_matlabroot():  # pragma: matlab
    r"""Test locate_matlabroot."""
    MatlabModelDriver.locate_matlabroot()


class TestMatlabModelDriver(base_class):  # pragma: matlab
    r"""Test parameters for MatlabModelDriver."""

    @pytest.fixture(scope="class")
    def language(self):
        r"""str: Language being tested."""
        return 'matlab'

    # @pytest.fixture(autouse=True)
    # def record_screen_session(self, instance, matlab_session_file):
    #     r"""Write the screen session to a file."""
    #     if instance.screen_session is None:  # pragma: debug
    #         logger.info("Matlab was not started by this test. Close any "
    #                     "existing Matlab sessions to test creation/removal.")
    #     else:
    #         with open(matlab_session_file, 'w') as f:
    #             f.write(instance.screen_session)
    #         instance.screen_session = None
    #         instance.started_matlab = False

    # @pytest.mark.skipif(not MatlabModelDriver._matlab_engine_installed,
    #                     reason="Matlab engine not installed.")
    # def test_z(self, instance, matlab_session_file):
    #     r"""Dummy test to stop matlab."""
    #     if os.path.isfile(matlab_session_file):
    #         with open(matlab_session_file, 'r') as f:
    #             session = f.read()
    #         os.remove(matlab_session_file)
    #         instance.screen_session = session
    #         instance.started_matlab = True
    #     else:  # pragma: debug
    #         logger.info("Skipping removal of Matlab session as the test did "
    #                     "not create it.")



import pytest
import os
from tests.examples import TestExample as base_class


class TestTimedPipeBase(base_class):
    r"""Base class for testing TimedPipe example with various comm types."""

    @pytest.fixture(scope="class")
    def example_name(self):
        r"""str: Name of example being tested."""
        return "timed_pipe"

    @pytest.fixture(scope="class")
    def env(self):
        r"""dict: Environment variables set for the test."""
        return {'PIPE_MSG_COUNT': '10',
                'PIPE_MSG_SIZE': '1024'}
        
    @pytest.fixture
    def results(self, env):
        r"""Result that should be found in output files."""
        siz = int(env['PIPE_MSG_COUNT']) * int(env['PIPE_MSG_SIZE'])
        res = '0' * siz
        return [res]

    @pytest.fixture
    def output_files(self, tempdir):
        r"""Output file."""
        return [os.path.join(tempdir, 'output_timed_pipe.txt')]

    @pytest.fixture(scope="class", autouse=True)
    def commtype(self, request, check_required_comms, change_default_comm,
                 language, running_service):
        r"""str: Comm used by the current test."""
        out = request.param
        if out in ['value', 'buffer', 'rmq_async']:
            pytest.skip("invalid commtype for integration")
        check_required_comms([out], language=language)
        with change_default_comm(out):
            if out == 'ipc':
                from yggdrasil.communication.IPCComm import (
                    ipcrm_queues, ipc_queues)
                qlist = ipc_queues()
                if qlist:  # pragma: debug
                    print('Existing queues:', qlist)
                    ipcrm_queues()
            if out == 'rest':
                with running_service('flask', partial_commtype='rest'):
                    yield out
            else:
                yield out



import pytest
try:
    import trimesh
except ImportError:  # pragma: debug
    trimesh = None
from tests.demos import DemoTstBase


@pytest.mark.skipif(trimesh is None, reason="Trimesh is not installed")
class TestFSPM2020Demo(DemoTstBase):

    @pytest.fixture(scope="class")
    def demo_name(self):
        r"""str: Name of demo being tested."""
        return 'fspm2020'

    runs = {'plant_v0': ('yamls/plant_v0.yml', ),
            'plant_v1': ('yamls/plant_v1.yml', 'yamls/light.yml'),
            'plant_v1_cpp': ('yamls/plant_v1.yml', 'yamls/light_cpp.yml'),
            'plant_v1_R': ('yamls/plant_v1.yml', 'yamls/light_R.yml'),
            'plant_v1_f90': ('yamls/plant_v1.yml',
                             'yamls/light_fortran.yml'),
            'plant_v2': ('yamls/plant_v2.yml', 'yamls/light.yml',
                         'yamls/roots.yml'),
            'call_split': ('yamls/plant_v2_split.yml', 'yamls/light.yml',
                           'yamls/roots.yml'),
            'simple_io': ('yamls/plant_output_mesh.yml',
                          'yamls/plant_io_mesh.yml')}



import pytest
from yggdrasil import platform
from yggdrasil.drivers.ExecutableModelDriver import ExecutableModelDriver


def test_execution_error():
    r"""Test errors raised during execution of invalid command."""
    with pytest.raises(RuntimeError):
        ExecutableModelDriver.run_executable(['invalid'])


def test_error_valgrind_strace():
    r"""Test error if both valgrind and strace set."""
    with pytest.raises(RuntimeError):
        ExecutableModelDriver('test', 'test',
                              with_strace=True, with_valgrind=True)


@pytest.mark.skipif(not platform._is_win, reason="Platform is not windows")
def test_error_valgrind_strace_windows():  # pragma: windows
    r"""Test error if strace or valgrind called on windows."""
    with pytest.raises(RuntimeError):
        ExecutableModelDriver('test', 'test', with_strace=True)
    with pytest.raises(RuntimeError):
        ExecutableModelDriver('test', 'test', with_valgrind=True)



import pytest
from tests.drivers.test_Driver import TestDriver as base_class
import os
import copy
import pprint
import shutil
import logging
from yggdrasil import platform, constants
from yggdrasil.drivers.ModelDriver import ModelDriver, remove_product
from yggdrasil.drivers.CompiledModelDriver import CompiledModelDriver
from yggdrasil.drivers.InterpretedModelDriver import InterpretedModelDriver


def test_remove_product():
    r"""Test remove_product."""
    test_file = os.path.join(os.path.dirname(__file__), 'remove_product_test.py')
    with open(test_file, 'w') as fd:
        fd.write('print(\'hello\')')
    try:
        with pytest.raises(RuntimeError):
            remove_product(test_file, check_for_source=True)
        with pytest.raises(RuntimeError):
            remove_product(os.path.dirname(test_file), check_for_source=True)
    finally:
        os.remove(test_file)


def test_ModelDriver_implementation():
    r"""Test that NotImplementedError raised for base class."""
    with pytest.raises(NotImplementedError):
        ModelDriver.language_executable()
    with pytest.raises(NotImplementedError):
        ModelDriver.executable_command(None)
    with pytest.raises(NotImplementedError):
        ModelDriver.is_library_installed(None)
    with pytest.raises(NotImplementedError):
        CompiledModelDriver.get_tool('compiler')
    with pytest.raises(NotImplementedError):
        InterpretedModelDriver.get_interpreter()


_models = sorted([x for x in (constants.LANGUAGES['interpreted']
                              + constants.LANGUAGES['dsl'])
                  if x not in ['matlab']] + ['executable'])


@pytest.mark.suite("models")
class TestModelDriver(base_class):
    r"""Test parameters for basic ModelDriver class."""

    _component_type = 'model'
    parametrize_language = _models

    @pytest.fixture(scope="class", autouse=True)
    def component_subtype(self, language):
        r"""Subtype of component being tested."""
        return language

    @pytest.fixture(scope="class")
    def language(self, request):
        r"""str: Language being tested."""
        return request.param

    @pytest.fixture(scope="class")
    def required_languages(self, language):
        r"""list: Languages required by the test."""
        return [language]

    @pytest.fixture(scope="class", autouse=True)
    def check_enabled(self, required_languages, check_required_languages):
        r"""Check if the language is enabled/disabled."""
        check_required_languages(required_languages)

    @pytest.fixture(scope="class")
    def source(self, python_class, scripts):
        r"""str: Source code to use for tests."""
        # if python_class.language is not None:
        out = scripts[python_class.language.lower()]
        if not isinstance(out, list):
            out = [out]
        return out

    @pytest.fixture(scope="class")
    def working_dir(self, source):
        r"""Working director."""
        return os.path.dirname(source[0])
    
    @pytest.fixture
    def instance_args(self, name, source, testing_options):
        r"""Arguments for a new instance of the tested class."""
        return tuple(
            [name, source + copy.deepcopy(
                testing_options.get('args', []))])

    @pytest.fixture(scope="class")
    def python_class_installed(self, python_class):
        r"""bool: True if the python class is installed."""
        return python_class.is_installed()

    @pytest.fixture
    def run_model_instance_kwargs(self):
        r"""dict: Additional keyword arguments that should be used in calls
        to run_model_instance"""
        return {}
    
    @pytest.fixture
    def run_model_instance(self, python_class, is_installed,
                           instance_args, instance_kwargs,
                           testing_options, run_model_instance_kwargs):
        r"""Create a driver for a model and run it."""
        if testing_options.get('requires_partner', False):
            pytest.skip(f"{python_class.language} requires partner model "
                        f"to run.")

        def run_model_instance_w(**kwargs):
            inst_kwargs = copy.deepcopy(instance_kwargs)
            inst_kwargs.update(kwargs)
            inst_kwargs.update(run_model_instance_kwargs)
            logger = logging.getLogger("yggdrasil")
            if logger.getEffectiveLevel() >= 20:
                inst_kwargs.setdefault('logging_level', 'INFO')
            drv = python_class(*instance_args, **inst_kwargs)
            getattr(drv, 'numeric_logging_level')
            drv.start()
            drv.wait(False)
            assert(not drv.errors)
            drv.cleanup()
        return run_model_instance_w

    @pytest.fixture(scope="class")
    def run_generated_code(self, python_class, is_installed):
        r"""Write and run generated code."""
        # def run_generated_code_w(lines, **kwargs):
        #     python_class.run_code(lines, **kwargs)
        return python_class.run_code

    @pytest.fixture(scope="class")
    def code_types(self, python_class, testing_options):
        if python_class.type_map is None:
            out = []
        else:
            out = copy.deepcopy(list(python_class.type_map.items()))
            if 'flag' not in python_class.type_map:
                out.append(('flag', python_class.type_map['boolean']))
        out += testing_options.get('code_types', [])
        for k, v in testing_options.get('replacement_code_types', {}).items():
            if k in out:
                out[out.index(k)] = v
        return out
            
    def test_executable_command(self, python_class):
        r"""Test error raise for invalid exec_type."""
        if python_class.executable_type != 'interpreter':
            pytest.skip("Only valid for languages with an interpreter")
        with pytest.raises(ValueError):
            python_class.executable_command([], exec_type='invalid')

    def test_is_installed(self, python_class, is_installed):
        r"""Assert that the tested model driver is installed."""
        assert(python_class.is_installed())

    def test_comm_installed(self, python_class):
        r"""Tests for getting installed comm while skipping config."""
        assert(python_class.is_comm_installed()
               == python_class.is_comm_installed(skip_config=True))
        assert(python_class.is_comm_installed(commtype='invalid',
                                              skip_config=True) is False)
        
    def test_language_version(self, python_class, is_installed):
        r"""Test language version."""
        assert(python_class.language_version())

    def test_python2language(self, python_class, testing_options,
                             nested_approx, pandas_equality_patch):
        r"""Test python2language."""
        for a, b in testing_options.get('python2language', []):
            assert(python_class.python2language(a) == nested_approx(b))
        
    def test_is_library_installed(self, python_class, testing_options):
        r"""Test is_library_installed for invalid library."""
        for x in testing_options.get('valid_libraries', []):
            assert(python_class.is_library_installed(x) is True)
        for x in testing_options.get('invalid_libraries', []):
            assert(python_class.is_library_installed(x) is False)
        
    def test_run_model(self, run_model_instance, testing_options):
        r"""Test running script used without debug."""
        if testing_options.get('requires_partner', False):
            pytest.skip("requires partner model to run")
        run_model_instance()

    @pytest.mark.skipif(platform._is_mac, reason="Valgrind slow on Mac")
    @pytest.mark.skipif(platform._is_win, reason="No valgrind on windows")
    @pytest.mark.skipif(shutil.which('valgrind') is None,
                        reason="Valgrind not installed.")
    def test_valgrind(self, working_dir, uuid, run_model_instance,
                      testing_options):
        r"""Test running with valgrind."""
        if testing_options.get('requires_partner', False):
            pytest.skip("requires partner model to run")
        valgrind_log = f"valgrind_log_{uuid.replace('-', '_')}.log"
        if working_dir:
            valgrind_log = os.path.join(working_dir, valgrind_log)
        try:
            run_model_instance(with_valgrind=True, with_strace=False,
                               valgrind_flags=['--leak-check=full',
                                               f'--log-file={valgrind_log}'])
        finally:
            if os.path.isfile(valgrind_log):
                os.remove(valgrind_log)
    
    @pytest.mark.skipif(platform._is_win or platform._is_mac,
                        reason="No strace on Windows or MacOS")
    @pytest.mark.skipif(shutil.which('strace') is None,
                        reason="strace not installed.")
    def test_strace(self, run_model_instance, testing_options):
        r"""Test running with strace."""
        if testing_options.get('requires_partner', False):
            pytest.skip("requires partner model to run")
        run_model_instance(with_valgrind=False, with_strace=True)
        
    # Tests for code generation
    def test_invalid_function_param(self, python_class, instance_kwargs,
                                    is_installed):
        r"""Test errors raise during class creation when parameters are invalid."""
        kwargs = copy.deepcopy(instance_kwargs)
        kwargs['name'] = 'test'
        kwargs['args'] = ['test']
        kwargs['function'] = 'invalid'
        kwargs['source_files'] = []
        if python_class.function_param is None:
            with pytest.raises(ValueError):
                python_class(**kwargs)
        else:
            kwargs['args'] = ['invalid']
            with pytest.raises(ValueError):
                python_class(**kwargs)
            kwargs['source_files'] = ['invalid']
            with pytest.raises(ValueError):
                python_class(**kwargs)
                               
    def test_get_native_type(self, python_class, code_types):
        r"""Test translation to native type."""
        for a, b in code_types:
            assert(python_class.get_native_type(datatype=a) == b)
            if not isinstance(a, dict):
                assert(
                    python_class.get_native_type(datatype={'type': a}) == b)
                if a in ['float', 'int', 'uint']:
                    assert(python_class.get_native_type(
                        datatype={'type': 'scalar', 'subtype': a}) == b)
                
    def test_write_declaration(self, python_class, code_types):
        r"""Test write_declaration for all supported native types."""
        if (((python_class.function_param is None)
             or ('declare' not in python_class.function_param))):
            pytest.skip("No declaration token")
        for a, b in code_types:
            python_class.write_declaration({'name': 'test', 'datatype': a})

    def test_write_model_wrapper(self, python_class):
        r"""Test writing a model based on yaml parameters."""
        if python_class.function_param is None:
            with pytest.raises(NotImplementedError):
                python_class.write_model_wrapper(None, None)
            with pytest.raises(NotImplementedError):
                python_class.write_model_recv(None, None)
            with pytest.raises(NotImplementedError):
                python_class.write_model_send(None, None)
        else:
            inputs = [
                {'name': 'a', 'datatype': 'bytes', 'outside_loop': True},
                {'name': 'b', 'datatype': {'type': 'int', 'precision': 64}},
                {'name': 'c', 'datatype': {'type': 'string', 'precision': 10}}]
            outputs = [
                {'name': 'y', 'datatype': {'type': 'float', 'precision': 32}},
                {'name': 'z', 'datatype': 'bytes', 'outside_loop': True},
                {'name': 'x', 'datatype': {'type': 'string', 'precision': 10}}]
            for iovar in [inputs, outputs]:
                for x in iovar:
                    x['name'] = 'test:' + x['name']
            python_class.write_model_wrapper('test', 'test',
                                             inputs=inputs,
                                             outputs=outputs)
            with pytest.raises(NotImplementedError):
                python_class.format_function_param('invalid_key')
        
    def test_write_executable(self, python_class):
        r"""Test writing an executable."""
        if python_class.function_param is None:
            with pytest.raises(NotImplementedError):
                python_class.write_executable(None)
        else:
            lines1 = python_class.write_executable('dummy',
                                                   prefix='dummy',
                                                   suffix='dummy')
            lines2 = python_class.write_executable(lines1)
            assert(lines1 == lines2)
            # Don't run this because it is invalid

    def test_error_code(self, python_class, is_installed):
        r"""Test that error is raised when code generates one."""
        if python_class.function_param is None:
            pytest.skip("Language not tokenized")
            return
        error_msg = 'Test error'
        lines = [python_class.function_param['error'].format(error_msg=error_msg)]
        with pytest.raises(RuntimeError):
            python_class.run_code(lines)

    @pytest.fixture(scope="class")
    def write_function_def(self, python_class, run_generated_code):
        r"""Write & run a function definition."""
        def write_function_def_w(inputs=None, outputs=None,
                                 outputs_in_inputs=None,
                                 declare_functions_as_var=None,
                                 guess_at_outputs_in_inputs=False, **kwargs):
            if declare_functions_as_var is None:
                declare_functions_as_var = False
            assert(inputs is not None)
            assert(outputs is not None)
            if outputs_in_inputs is None:
                outputs_in_inputs = python_class.outputs_in_inputs
            flag_var = {'name': 'flag',
                        'datatype': 'flag',
                        'value': python_class.function_param['true']}
            function_contents = []
            if len(inputs) == len(outputs):
                for i, o in zip(inputs, outputs):
                    function_contents += python_class.write_assign_to_output(
                        o['name'], i['name'],
                        outputs_in_inputs=outputs_in_inputs)
                    function_contents += python_class.write_print_output_var(
                        o['name'], in_inputs=outputs_in_inputs)
            output_var = python_class.prepare_output_variables(
                outputs, in_inputs=python_class.outputs_in_inputs,
                in_definition=True)
            if not python_class.types_in_funcdef:
                kwargs['outputs'] = outputs
            definition = python_class.write_function_def(
                'test_function', inputs=inputs, output_var=output_var,
                function_contents=function_contents,
                flag_var=flag_var, outputs_in_inputs=outputs_in_inputs,
                print_inputs=True, print_outputs=True, **kwargs)
            # Add second definition to test ability to locate specific
            # function in the presence of others
            definition.append('')
            definition += python_class.write_function_def(
                'test_function_decoy', inputs=inputs,
                outputs=outputs, flag_var=flag_var,
                function_contents=function_contents,
                outputs_in_inputs=outputs_in_inputs,
                skip_interface=True)
            parsed = None
            try:
                kwargs = dict(contents='\n'.join(definition),
                              expected_outputs=outputs,
                              outputs_in_inputs=outputs_in_inputs)
                if guess_at_outputs_in_inputs:
                    kwargs.pop('expected_outputs')
                    kwargs.pop('outputs_in_inputs')
                parsed = python_class.parse_function_definition(
                    None, 'test_function', **kwargs)
                assert(len(parsed.get('inputs', [])) == len(inputs))
                assert(len(parsed.get('outputs', [])) == len(outputs))
            except BaseException:  # pragma: debug
                pprint.pprint(definition)
                if parsed:
                    pprint.pprint(parsed)
                    print("Expected inputs:")
                    pprint.pprint(inputs)
                    print("Expected outputs:")
                    pprint.pprint(outputs)
                raise
            if inputs:
                for xp, x0 in zip(parsed['inputs'], inputs):
                    assert(xp['name'] == x0['name'])
                    x0.update(xp)
            if outputs:
                for xp, x0 in zip(parsed['outputs'], outputs):
                    assert(xp['name'] == x0['name'])
                    x0.update(xp)
            # Lines required to set up the function call
            lines = []
            if 'declare' in python_class.function_param:
                for x in inputs + outputs:
                    lines += python_class.write_declaration(x)
                if outputs_in_inputs:
                    lines += python_class.write_declaration(flag_var)
                if declare_functions_as_var:
                    if outputs_in_inputs:
                        lines += python_class.write_declaration(
                            dict(flag_var, name='test_function'))
                    elif len(outputs) > 0:
                        lines += python_class.write_declaration(
                            dict(outputs[0], name='test_function'))
            for x in inputs:
                lines.append(python_class.format_function_param(
                    'assign', **x))
            if outputs_in_inputs:
                lines.append(python_class.format_function_param(
                    'assign', **flag_var))
            lines += python_class.write_function_call(
                'test_function', flag_var=flag_var,
                inputs=inputs, outputs=outputs,
                outputs_in_inputs=outputs_in_inputs)
            run_generated_code(lines,
                               function_definitions=definition)
        return write_function_def_w

    def test_write_function_def(self, python_class, write_function_def,
                                testing_options):
        r"""Test writing and running a function definition."""
        if python_class.function_param is None:
            with pytest.raises(NotImplementedError):
                python_class.write_function_def(None)
        else:
            for kws in testing_options.get('write_function_def_params', []):
                write_function_def(**kws)
            
    def test_write_if_block(self, python_class, run_generated_code):
        r"""Test writing an if block."""
        if python_class.function_param is None:
            with pytest.raises(NotImplementedError):
                python_class.write_if_block(None, None)
        else:
            lines = []
            if 'declare' in python_class.function_param:
                lines += python_class.write_declaration(
                    {'name': 'x', 'type': 'integer'})
            cond = [python_class.function_param['true'],
                    python_class.function_param['false']]
            block_contents = [
                python_class.function_param['assign'].format(
                    name='x', value='1'),
                python_class.function_param['assign'].format(
                    name='x', value='2')]
            else_contents = python_class.function_param['assign'].format(
                name='x', value='-1')
            lines += python_class.write_if_block(
                cond, block_contents,
                else_block_contents=else_contents)
            run_generated_code(lines)

    def test_write_for_loop(self, python_class, run_generated_code):
        r"""Test writing a for loop."""
        if python_class.function_param is None:
            with pytest.raises(NotImplementedError):
                python_class.write_for_loop(None, None, None, None)
        else:
            lines = []
            if 'declare' in python_class.function_param:
                lines += python_class.write_declaration(
                    {'name': 'i', 'type': 'integer'})
                lines += python_class.write_declaration(
                    {'name': 'x', 'type': 'integer'})
            loop_contents = python_class.function_param['assign'].format(
                name='x', value='i')
            lines += python_class.write_for_loop('i', 1, 2, loop_contents)
            run_generated_code(lines)

    def test_write_while_loop(self, python_class, run_generated_code):
        r"""Test writing a while loop."""
        if python_class.function_param is None:
            with pytest.raises(NotImplementedError):
                python_class.write_while_loop(None, None)
        else:
            lines = []
            cond = python_class.function_param['true']
            loop_contents = python_class.function_param.get('break', 'break')
            lines += python_class.write_while_loop(cond, loop_contents)
            run_generated_code(lines)

    def test_write_try_except(self, python_class, run_generated_code,
                              testing_options):
        r"""Test writing a try/except block."""
        if (((python_class.function_param is None)
             or ('try_begin' not in python_class.function_param))):
            with pytest.raises(NotImplementedError):
                python_class.write_try_except(None, None)
        else:
            lines = []
            try_contents = python_class.function_param['error'].format(
                error_msg='Dummy error')
            except_contents = python_class.function_param['print'].format(
                message='Dummy message')
            lines += python_class.write_try_except(
                try_contents, except_contents,
                **testing_options.get('write_try_except_kwargs', {}))
            run_generated_code(lines)

    def test_cleanup_dependencies(self, python_class):
        r"""Test cleanup_dependencies method."""
        # Run twice to ensure things work even if it has been cleand up
        python_class.cleanup_dependencies()
        python_class.cleanup_dependencies()

    def test_split_line(self, python_class, testing_options):
        r"""Test split_line."""
        if python_class.function_param is None:
            return
        for line, kwargs, splits in testing_options.get('split_lines', []):
            assert(python_class.split_line(line, **kwargs) == splits)

    def test_install_model_dependencies(self, python_class, testing_options):
        r"""Test install_model_dependencies."""
        deps = testing_options.get('deps', [])
        python_class.install_model_dependencies(deps, always_yes=True)
        with pytest.raises(NotImplementedError):
            python_class.install_dependency(
                'invalid', package_manager='invalid')



import os
import re
import copy
from yggdrasil import platform
from yggdrasil.drivers.CModelDriver import (
    CCompilerBase, CModelDriver, GCCCompiler, ClangCompiler, MSVCCompiler,
    ClangLinker)


class CPPCompilerBase(CCompilerBase):
    r"""Base class for C++ compilers."""
    languages = ['c++']
    default_executable_env = 'CXX'
    default_flags_env = 'CXXFLAGS'
    cpp_std = 'c++14'
    search_path_flags = ['-E', '-v', '-xc++', '/dev/null']
    default_linker = None
    default_executable = None

    @classmethod
    def add_standard_flag(cls, flags):
        r"""Add a standard flag to the list of flags.

        Args:
            flags (list): Compilation flags.

        """
        std_flag = None
        for i, a in enumerate(flags):
            if a.startswith('-std='):
                std_flag = i
                break
        if std_flag is None:
            flags.append('-std=%s' % cls.cpp_std)
        return flags


class GPPCompiler(CPPCompilerBase, GCCCompiler):
    r"""Interface class for G++ compiler/linker."""
    toolname = 'g++'
    aliases = ['gnu-c++']

    @classmethod
    def get_flags(cls, skip_standard_flag=False, **kwargs):
        r"""Get a list of compiler flags.

        Args:
            skip_standard_flag (bool, optional): If True, the C++ standard flag
                will not be added. Defaults to False.
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        Returns:
            list: Compiler flags.

        """
        out = super(GPPCompiler, cls).get_flags(**kwargs)
        # Add standard library flag
        if not skip_standard_flag:
            out = cls.add_standard_flag(out)
        return out


class ClangPPCompiler(CPPCompilerBase, ClangCompiler):
    r"""Interface class for clang++ compiler."""
    toolname = 'clang++'
    default_linker = 'clang++'
    # Set to False since ClangLinker has its own class to handle
    # conflict between versions of clang and ld.
    is_linker = False

    @staticmethod
    def before_registration(cls):
        r"""Operations that should be performed to modify class attributes prior
        to registration including things like platform dependent properties and
        checking environment variables for default settings.
        """
        if platform._is_win:  # pragma: windows
            cls.default_executable = 'clang'
        CPPCompilerBase.before_registration(cls)

    @classmethod
    def get_flags(cls, skip_standard_flag=False, **kwargs):
        r"""Get a list of compiler flags.

        Args:
            skip_standard_flag (bool, optional): If True, the C++ standard flag
                will not be added. Defaults to False.
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        Returns:
            list: Compiler flags.

        """
        out = super(ClangPPCompiler, cls).get_flags(**kwargs)
        # Add standard library flag
        if not skip_standard_flag:
            out = cls.add_standard_flag(out)
        return out
        
    @classmethod
    def get_executable_command(cls, args, skip_flags=False, unused_kwargs=None,
                               **kwargs):
        r"""Determine the command required to run the tool using the specified
        arguments and options.

        Args:
            args (list): The arguments that should be passed to the tool. If
                skip_flags is False, these are treated as input files that will
                be used by the tool.
            **kwargs: Additional keyword arguments will be passed to the parent
                class's method.

        Returns:
            str: Output to stdout from the command execution.

        """
        if platform._is_win:  # pragma: windows
            for a in args:
                if a.endswith('.c'):
                    kwargs['skip_standard_flag'] = True
                    break
        return super(ClangPPCompiler, cls).get_executable_command(args, **kwargs)


class MSVCPPCompiler(CPPCompilerBase, MSVCCompiler):
    r"""Inteface class for MSVC compiler when compiling C++."""
    toolname = 'cl++'
    default_linker = MSVCCompiler.default_linker
    default_archiver = MSVCCompiler.default_archiver
    default_executable = MSVCCompiler.default_executable
    search_path_flags = None
    dont_create_linker = True
    
    @staticmethod
    def before_registration(cls):
        r"""Operations that should be performed to modify class attributes prior
        to registration including things like platform dependent properties and
        checking environment variables for default settings.
        """
        return MSVCCompiler.before_registration(cls)


class ClangPPLinker(ClangLinker):
    r"""Interface class for clang++ linker (calls to ld)."""
    toolname = ClangPPCompiler.toolname
    languages = ClangPPCompiler.languages
    default_executable = ClangPPCompiler.default_executable
    toolset = ClangPPCompiler.toolset


class CPPModelDriver(CModelDriver):
    r"""Class for running C++ models."""
                
    _schema_subtype_description = ('Model is written in C++.')
    language = 'c++'
    language_ext = ['.cpp', '.CPP', '.cxx', '.C', '.c++', '.cc', '.cp', '.tcc',
                    '.hpp', '.HPP', '.hxx', '.H', '.h++', '.hh', '.hp', '.h']
    language_aliases = ['cpp', 'cxx']
    base_languages = ['c']
    interface_library = 'ygg++'
    # To prevent inheritance
    default_compiler = None
    default_linker = None
    interface_map = {
        'import': '#include "YggInterface.hpp"',
        'input': 'YggInput {channel_obj}("{channel_name}")',
        'output': 'YggOutput {channel_obj}("{channel_name}")',
        'server': (
            'YggRpcServer {channel_obj}("{channel_name}", '
            '{datatype_in}, {datatype_out})'),
        'client': (
            'YggRpcClient {channel_obj}("{channel_name}", '
            '{datatype_out}, {datatype_in})'),
        'timesync': 'YggTimesync {channel_obj}("{channel_name}", "{time_units}")',
        'send': 'flag = {channel_obj}.send({nargs}, {outputs})',
        'recv': 'flag = {channel_obj}.recv({nargs}, {input_refs})',
        'call': 'flag = {channel_obj}.call({nargs}, {outputs}, {input_refs})',
    }
    function_param = dict(
        CModelDriver.function_param,
        input='YggInput {channel}(\"{channel_name}\", {channel_type});',
        output='YggOutput {channel}(\"{channel_name}\", {channel_type});',
        recv_heap='{channel}.recvRealloc',
        recv_stack='{channel}.recv',
        recv_function='{channel}.recvRealloc',
        send_function='{channel}.send',
        exec_prefix=('#include <iostream>\n'
                     '#include <exception>\n'),
        print_generic='std::cout << {object} << std::endl << std::flush;',
        error='throw \"{error_msg}\";',
        try_begin='try {',
        try_error_type='const std::exception&',
        try_except='}} catch ({error_type} {error_var}) {{',
        function_def_regex=(
            r'(?P<flag_type>.+?)\s*{function_name}\s*'
            r'\((?P<inputs>(?:[^{{\&])*?)'
            r'(?:,\s*(?P<outputs>'
            r'(?:\s*(?:[^\s\&]+)'
            r'(?:(?:\&\s+)|(?:\s+(?:\()?\&))'
            r'(?:[^{{])+)+))?\)\s*\{{'
            r'(?P<body>(?:.*?\n?)*?)'
            r'(?:(?:return +(?P<flag_var>.+?)?;(?:.*?\n?)*?\}})'
            r'|(?:\}}))'),
        outputs_def_regex=(
            r'\s*(?P<native_type>(?:[^\s])+)(\s+)?'
            r'(\()?(?P<ref>\&)(?(1)(?:\s*)|(?:\s+))'
            r'(?P<name>.+?)(?(2)(?:\)|(?:)))(?P<shape>(?:\[.+?\])+)?\s*(?:,|$)(?:\n)?'),
        vector_regex=r'(?:std\:\:)?vector\<\s*(?P<type>.*)\s*\>')
    include_arg_count = True
    include_channel_obj = False
    dont_declare_channel = True
    
    @staticmethod
    def after_registration(cls, **kwargs):
        r"""Operations that should be performed to modify class attributes after
        registration."""
        if cls.default_compiler is None:
            if platform._is_linux:
                cls.default_compiler = 'g++'
            elif platform._is_mac:
                cls.default_compiler = 'clang++'
            elif platform._is_win:  # pragma: windows
                cls.default_compiler = 'cl'
        cls.function_param['print'] = 'std::cout << "{message}" << std::endl;'
        CModelDriver.after_registration(cls, **kwargs)
        if kwargs.get('second_pass', False):
            return
        internal_libs = copy.deepcopy(cls.internal_libraries)
        internal_libs[cls.interface_library] = internal_libs.pop(
            CModelDriver.interface_library)
        internal_libs[cls.interface_library]['source'] = os.path.join(
            cls.get_language_dir(),
            os.path.splitext(os.path.basename(
                internal_libs[cls.interface_library]['source']))[0]
            + cls.language_ext[0])
        internal_libs[cls.interface_library]['include_dirs'].append(
            cls.get_language_dir())
        internal_libs[cls.interface_library]['language'] = cls.language
        cls.internal_libraries = internal_libs

    @classmethod
    def set_env_class(cls, **kwargs):
        r"""Set environment variables that are instance independent.

        Args:
            **kwargs: Additional keyword arguments are passed to the parent
                class's method and update_ld_library_path.

        Returns:
            dict: Environment variables for the model process.

        """
        out = super(CPPModelDriver, cls).set_env_class(**kwargs)
        out = CModelDriver.update_ld_library_path(out, **kwargs)
        return out
        
    @classmethod
    def write_try_except(cls, try_contents, except_contents, error_var='e',
                         error_type=None, **kwargs):
        r"""Return the lines required to complete a try/except block.

        Args:
            try_contents (list): Lines of code that should be executed inside
                the try block.
            except_contents (list): Lines of code that should be executed inside
                the except block.
            error_var (str, optional): Name of variable where the caught error
                should be stored. Defaults to 'e'. If '...', the catch clause
                will catch all errors, but there will not be a name error.
            error_type (str, optional): Name of error type that should be caught.
                If not provided, defaults to None and will be set based on the
                class function_param entry for 'try_error_type'. If '...', the
                catch clause will catch all errors and error_var will be
                ignored.
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        Returns:
            Lines of code perfoming a try/except block.

        """
        if (error_type == '...') or (error_var == '...'):
            error_type = ''
            error_var = '...'
        kwargs.update(error_var=error_var, error_type=error_type)
        return super(CPPModelDriver, cls).write_try_except(
            try_contents, except_contents, **kwargs)

    @classmethod
    def finalize_function_io(cls, direction, x):
        r"""Finalize info for an input/output channel following function
        parsing.

        Args:
            direction (str): Direction of channel ('input' or 'output')
            x (dict): Channel info.

        """
        if direction == 'input':
            for v in x['vars']:
                grp_vect = cls.is_vector(v)
                if grp_vect:
                    v['ptr_var'] = dict(v, name=(v['name'] + '_ptr'))
                    v['ptr_var'].pop('native_type')
        elif direction == 'output':
            for v in x['vars']:
                if (not v.get('length_var', False)) and cls.is_vector(v):
                    v['length_var'] = v['name'] + '.size()'
                    v['ptr_var'] = v['name'] + '.data()'
        super(CPPModelDriver, cls).finalize_function_io(direction, x)
        if direction == 'input':
            for v in x['vars']:
                if 'ptr_var' in v:
                    v['ptr_var']['length_var'] = v['length_var']
        
    @classmethod
    def is_vector(cls, var):
        r"""Determine if a variable uses a vector.

        Args:
            var (dict): Variable.

        Returns:
            bool: True if it is a vector, False otherwise.

        """
        if isinstance(var, dict) and ('vector' in var.get('native_type', '')):
            return re.fullmatch(cls.function_param['vector_regex'],
                                var['native_type'])
        return False
    
    @classmethod
    def requires_length_var(cls, var):
        r"""Determine if a variable requires a separate length variable.

        Args:
            var (dict): Dictionary of variable properties.

        Returns:
            bool: True if a length variable is required, False otherwise.

        """
        if cls.is_vector(var):
            return True
        return super(CPPModelDriver, cls).requires_length_var(var)

    @classmethod
    def write_declaration(cls, var, **kwargs):
        r"""Return the lines required to declare a variable with a certain
        type.

        Args:
            var (dict, str): Name or information dictionary for the variable
                being declared.
            **kwargs: Additional keyword arguments are passed to the
                parent class's method.

        Returns:
            list: The lines declaring the variable.

        """
        out = []
        if 'ptr_var' in var:
            var_copy = copy.deepcopy(var)
            if isinstance(var['ptr_var'], dict):
                out += super(CPPModelDriver, cls).write_declaration(
                    var['ptr_var'], **kwargs)
            out += super(CModelDriver, cls).write_declaration(var_copy,
                                                              **kwargs)
            return out
        out += super(CPPModelDriver, cls).write_declaration(var, **kwargs)
        return out
        
    @classmethod
    def write_model_recv(cls, channel, recv_var, **kwargs):
        r"""Write a model receive call include checking the return flag.

        Args:
            channel (str): Name of variable that the channel being received
                from was stored in.
            recv_var (dict, list): Information of one or more variables that
                receieved information should be stored in.
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        Returns:
            list: Lines required to carry out a receive call in this language.

        """
        out_after = []
        if not isinstance(recv_var, str):
            recv_var_par = cls.channels2vars(recv_var)
            allows_realloc = [cls.allows_realloc(v)
                              for v in recv_var_par]
            is_vector = [cls.is_vector(v) for v in recv_var_par]
            if any(is_vector):
                if all(allows_realloc):
                    kwargs.setdefault('alt_recv_function',
                                      cls.function_param['recv_heap'])
                else:  # pragma: debug
                    # kwargs.setdefault('alt_recv_function',
                    #                   cls.function_param['recv_stack'])
                    raise RuntimeError("Mixing vectors with stack allocated "
                                       "arrays is not get supported.")
                new_recv_var_par = []
                for i, v in enumerate(recv_var_par):
                    if cls.allows_realloc(v) and cls.is_vector(v):
                        assert(v.get('ptr_var', False))
                        out_after.append(
                            '{var}.assign({ptr_var}, '
                            '{ptr_var} + {len_var});'.format(
                                var=v['name'], ptr_var=v['ptr_var']['name'],
                                len_var=v['length_var']['name']))
                        v = v['ptr_var']
                    new_recv_var_par.append(v)
                recv_var_par = new_recv_var_par
                recv_var = cls.prepare_output_variables(
                    recv_var_par, in_inputs=cls.outputs_in_inputs,
                    for_yggdrasil=True)
        out = super(CPPModelDriver, cls).write_model_recv(
            channel, recv_var, **kwargs)
        return out + out_after

    @classmethod
    def write_model_send(cls, channel, send_var, **kwargs):
        r"""Write a model send call include checking the return flag.

        Args:
            channel (str): Name of variable that the channel being sent to
                was stored in.
            send_var (dict, list): Information on one or more variables
                containing information that will be sent.
            flag_var (str, optional): Name of flag variable that the flag
                should be stored in. Defaults to 'flag',
            allow_failure (bool, optional): If True, the returned lines will
                call a break if the flag is False. Otherwise, the returned
                lines will issue an error. Defaults to False.

        Returns:
            list: Lines required to carry out a send call in this language.

        """
        send_var_str = send_var
        out_before = []
        if not isinstance(send_var, str):
            send_var_par = []
            for v in cls.channels2vars(send_var):
                if cls.is_vector(v):
                    send_var_par += [v['ptr_var'], v['length_var']]
                else:
                    send_var_par.append(v)
            send_var_str = cls.prepare_input_variables(
                send_var_par, for_yggdrasil=True)
        out = super(CPPModelDriver, cls).write_model_send(
            channel, send_var_str, **kwargs)
        return out_before + out
    
    @classmethod
    def output2input(cls, var, in_definition=True):
        r"""Perform conversion necessary to turn an output variable
        into an corresponding input that can be used to format a
        function definition.

        Args:
            var (dict): Variable definition.
            in_definition (bool, optional): If True, the returned
                dictionary corresponds to an input variable in a
                function definition. If False, the returned value
                will correspond to an input to a function. Defaults to
                True.

        Returns:
            dict: Updated variable definition.

        """
        out = super(CModelDriver, cls).output2input(var)
        if isinstance(var, dict):
            if in_definition:
                out = dict(out, name='&' + out['name'])
                if ((('shape' in out.get('datatype', {}))
                     or ('length' in out.get('datatype', {})))):
                    out['name'] = '(%s)' % out['name']
            else:
                if not (out.get('ref', False)
                        or out.get('is_length_var', False)):
                    out = dict(out, name='&' + out['name'])
        return out

    @classmethod
    def get_json_type(cls, native_type):
        r"""Get the JSON type from the native language type.

        Args:
            native_type (str): The native language type.

        Returns:
            str, dict: The JSON type.

        """
        regex_vect = r'(?:std\:\:)?vector\<\s*(?P<type>.*)\s*\>'
        grp_vect = re.fullmatch(regex_vect, native_type)
        if grp_vect:
            out = cls.get_json_type(grp_vect.group('type'))
            if out['type'] == 'scalar':
                out['type'] = '1darray'
            else:  # pragma: debug
                raise ValueError("Type not currently supported: %s -> %s"
                                 % (native_type, out))
        else:
            out = super(CPPModelDriver, cls).get_json_type(native_type)
        return out

    @classmethod
    def get_testing_options(cls, **kwargs):
        r"""Method to return a dictionary of testing options for this class.

        Args:
            **kwargs: Additional keyword arguments are passed to the parent
                class.

        Returns:
            dict: Dictionary of variables to use for testing. Key/value pairs:
                kwargs (dict): Keyword arguments for driver instance.
                deps (list): Dependencies to install.

        """
        from yggdrasil import tools
        out = super(CPPModelDriver, cls).get_testing_options(**kwargs)
        out['deps'] = ["cmake",
                       {"package_manager": "pip", "package": "pyyaml",
                        "arguments": "-v"},
                       {"package": "cmake", "arguments": "-v"}]
        if platform._is_win:  # pragma: windows
            if not tools.get_conda_prefix():
                out['deps'].append({"package_manager": "vcpkg",
                                    "package": "czmq"})
            else:
                out['deps'].append('doxygen')
        out['write_try_except_kwargs'] = {'error_type': '...'}
        out['kwargs'].setdefault('compiler_flags', [])
        out['kwargs']['compiler_flags'].append('-std=c++11')
        return out



import pytest
from tests import TestClassBase as base_class
from tests.drivers.test_ModelDriver import TestModelDriver as model_base_class
import os
import copy
import shutil
from yggdrasil import platform, constants
from yggdrasil.config import ygg_cfg
from yggdrasil.drivers import CompiledModelDriver
from yggdrasil.components import import_component


def test_get_compatible_tool():
    r"""Test get_compatible_tool when default provided."""
    with pytest.raises(ValueError):
        CompiledModelDriver.get_compatible_tool('invalid', 'compiler', 'c')
    assert(CompiledModelDriver.get_compatible_tool(
        'invalid', 'compiler', 'c', default=None) is None)


def test_get_compilation_tool_registry():
    r"""Test errors raised by get_compilation_tool_registry."""
    with pytest.raises(ValueError):
        CompiledModelDriver.get_compilation_tool_registry('invalid')

    
def test_find_compilation_tool():
    r"""Test errors raised by find_compilation_tool."""
    with pytest.raises(RuntimeError):
        CompiledModelDriver.find_compilation_tool('archiver', 'cmake')


def test_get_alternate_class():
    r"""Test get_alternate_class."""
    import_component('model', subtype='c')
    gcc = CompiledModelDriver.get_compilation_tool('compiler', 'gcc')
    gcc.get_alternate_class(toolname='clang')
    

def test_get_compilation_tool():
    r"""Test get_compilation_tool for different name variations."""
    from yggdrasil.drivers.CModelDriver import CModelDriver
    if CModelDriver.is_language_installed():
        tooltype = 'compiler'
        out = CModelDriver.get_tool('compiler').__class__
        toolname = out.toolname.lower()
        toolpath = os.path.join('somedir', toolname)
        toolfile = toolpath + '.exe'
        vals = [toolpath, toolfile]
        if platform._is_win:
            vals += [toolname.upper(), toolfile.upper()]
        for v in vals:
            assert(CompiledModelDriver.get_compilation_tool(tooltype, v) == out)
        with pytest.raises(ValueError):
            CompiledModelDriver.get_compilation_tool('compiler', 'invalid')
    else:
        with pytest.raises(NotImplementedError):
            CModelDriver.get_tool('compiler')
        assert(CModelDriver.get_tool(
            'compiler', default='invalid') == 'invalid')
    assert(CompiledModelDriver.get_compilation_tool('compiler', 'invalid',
                                                    default='invalid')
           == 'invalid')


def test_CompilationToolBase():
    r"""Test error in CompilationToolBase."""
    with pytest.raises(RuntimeError):
        CompiledModelDriver.CompilationToolBase(invalid='invalid')


class DummyCompiler(CompiledModelDriver.CompilerBase):
    r"""Dummy test class."""
    _dont_register = True
    toolname = 'dummy12345'
    languages = ['dummy']
    search_path_envvar = ['PATH']
    _language_ext = ['.c']
    default_linker = False
    default_archiver = None
    combine_with_linker = True
    compile_only_flag = None
    linker_attributes = {'_dont_register': True}


class TestCompilationTool(base_class):
    r"""Test class for compilation tools."""

    _mod = 'yggdrasil.drivers.CompiledModelDriver'
    _cls = 'CompilationToolBase'

    def test_append_flags(self, python_class):
        r"""Test append_flags."""
        with pytest.raises(ValueError):
            python_class.append_flags([], '-T', 'bye', invalid='invalid')
        with pytest.raises(ValueError):
            python_class.append_flags(['-Thello'], '-T%s', 'bye',
                                      no_duplicates=True)
        kws_list = [((['a', 'b', 'c'], '0', '1'), {'prepend': True},
                     ['0', '1', 'a', 'b', 'c']),
                    ((['a', 'b', 'c'], '0', '1'), {'position': -1},
                     ['a', 'b', 'c', '0', '1']),
                    ((['a', 'b', 'c'], '0', '1'), {'position': -2},
                     ['a', 'b', '0', '1', 'c'])]
        for (out, key, val), kws, res in kws_list:
            python_class.append_flags(out, key, val, **kws)
            assert(out == res)

    def test_create_flag(self, python_class):
        r"""Test create_flag."""
        test_args = [({'key': '-T%s'}, ['a', 'b', 'c'], ['-Ta', '-Tb', '-Tc']),
                     ('-set-this', True, ['-set-this']),
                     ('-set-this', False, []),
                     ('-set-this', None, [])]
        for (key, val, out) in test_args:
            assert(python_class.create_flag(key, val) == out)

    def test_not_implemented(self):
        r"""Test raising of NotImplementedErrors for incomplete classes."""
        pass

    def test_get_flags(self, python_class):
        r"""Test get_flags."""
        assert(python_class.get_flags(flags='hello') == ['hello'])

    def test_get_search_path(self, class_name, python_class):
        r"""Test get_search_path."""
        if class_name == 'CompilationToolBase':
            with pytest.raises(NotImplementedError):
                python_class.get_search_path()
        else:
            python_class.get_search_path(libtype='include')
            python_class.get_search_path(libtype='shared')
            python_class.get_search_path(libtype='static')
            
    def test_get_executable_command(self, python_class):
        r"""Test get_executable_command."""
        if python_class.toolname is None:
            with pytest.raises(NotImplementedError):
                python_class.get_executable_command([])
            

class TestDummyCompiler(TestCompilationTool):
    r"""Test class for DummyCompiler."""

    _mod = 'tests.drivers.test_CompiledModelDriver'
    _cls = 'DummyCompiler'

    @pytest.fixture
    def instance_kwargs(self):
        r"""Keyword arguments for a new instance of the tested class."""
        return dict(linker=False)

    def test_call(self, python_class):
        r"""Test call."""
        out = 'test123'
        assert(not shutil.which(python_class.toolname))
        assert(not (os.path.isfile(out) or os.path.isdir(out)))
        with pytest.raises(RuntimeError):
            python_class.call('args', out=out, dont_link=True)

    def test_linker(self, python_class):
        r"""Test linker."""
        assert(python_class.linker() is False)

    def test_archiver(self, python_class):
        r"""Test archiver."""
        with pytest.raises(RuntimeError):
            python_class.archiver()
        
    def test_get_flags(self, python_class):
        r"""Test get_flags."""
        assert(python_class.get_flags(flags='hello', libtype='object')
               == ['hello', '-DWITH_YGGDRASIL'])
        
    def test_get_executable_command(self, python_class):
        r"""Test get_executable_command."""
        python_class.get_executable_command([], dont_link=True)


class TestCompiledModelDriver(model_base_class):
    r"""Test runner for CompiledModelDriver."""

    parametrize_language = constants.LANGUAGES['compiled']

    @pytest.fixture(scope="class")
    def compiler(self, python_class):
        r"""Compiler for the class."""
        return python_class.get_tool('compiler')
        
    @pytest.fixture
    def instance_args(self, name, source, testing_options, compiler,
                      is_installed):
        r"""Arguments for a new instance of the tested class."""
        return tuple(
            [name, ([compiler.get_output_file(source[0])]
                    + copy.deepcopy(
                        testing_options.get('args', [])))])

    @pytest.fixture
    def instance_kwargs(self, testing_options, timeout, working_dir,
                        polling_interval, namespace, source):
        r"""Keyword arguments for a new instance of the tested class."""
        return dict(
            copy.deepcopy(
                testing_options.get('kwargs', {})),
            yml={'working_dir': working_dir},
            timeout=timeout, sleeptime=polling_interval,
            namespace=namespace, source_files=source)
    
    @pytest.fixture
    def run_model_instance_kwargs(self):
        r"""dict: Additional keyword arguments that should be used in calls
        to run_model_instance"""
        return {'skip_compile': False}
        
    def test_build(self, python_class):
        r"""Test building libraries as a shared/static library or object files."""
        # Finish on the default libtype
        order = ['shared', 'object', 'static']
        order.remove(CompiledModelDriver._default_libtype)
        order.append(CompiledModelDriver._default_libtype)
        for libtype in ['shared', 'object', 'static']:
            python_class.compile_dependencies(
                libtype=libtype, overwrite=True)
            if libtype == 'shared':
                python_class.compile_dependencies(
                    libtype=libtype, overwrite=False)
            python_class.cleanup_dependencies(libtype=libtype)

    def test_get_tool(self, python_class):
        r"""Test other methods of calling get_tool."""
        python_class.get_tool('compiler', return_prop='name')
        python_class.get_tool('compiler', return_prop='flags')
        with pytest.raises(ValueError):
            python_class.get_tool('compiler', return_prop='invalid')

    def test_get_dependency_info(self, python_class):
        r"""Test get_dependency_info."""
        dep_list = (
            python_class.get_dependency_order(
                python_class.interface_library)
            + list(python_class.external_libraries.keys()))
        for dep in dep_list:
            python_class.get_dependency_info(dep, default='default')
        with pytest.raises(KeyError):
            python_class.get_dependency_info('invalid')
        assert(python_class.get_dependency_info('invalid', default='default')
               == 'default')

    def test_get_dependency_source(self, python_class):
        r"""Test get_dependency_source."""
        dep_list = (
            python_class.get_dependency_order(
                python_class.interface_library)
            + list(python_class.external_libraries.keys()))
        for dep in dep_list:
            python_class.get_dependency_source(dep, default='default')
        with pytest.raises(ValueError):
            python_class.get_dependency_source('invalid')
        assert(python_class.get_dependency_source(__file__) == __file__)
        assert(python_class.get_dependency_source(
            'invalid', default='default') == 'default')

    def test_get_dependency_object(self, python_class):
        r"""Test get_dependency_object."""
        dep_list = (
            python_class.get_dependency_order(
                python_class.interface_library)
            + list(python_class.external_libraries.keys()))
        for dep in dep_list:
            python_class.get_dependency_object(dep, default='default')
        with pytest.raises(ValueError):
            python_class.get_dependency_object('invalid')
        assert(python_class.get_dependency_object(__file__) == __file__)
        assert(python_class.get_dependency_object(
            'invalid', default='default') == 'default')

    def test_get_dependency_library(self, python_class):
        r"""Test get_dependency_library."""
        with pytest.raises(ValueError):
            python_class.get_dependency_library('invalid', libtype='invalid')
        for dep, info in python_class.external_libraries.items():
            libtype_orig = info.get('libtype', None)
            if libtype_orig not in ['static', 'shared']:
                continue
            if libtype_orig == 'static':  # pragma: no cover
                libtype = 'shared'
            else:
                libtype = 'static'
            with pytest.raises(ValueError):
                python_class.get_dependency_library(dep, libtype=libtype)
        with pytest.raises(ValueError):
            python_class.get_dependency_library('invalid')
        assert(python_class.get_dependency_library(__file__) == __file__)
        assert(python_class.get_dependency_library(
            'invalid', default='default') == 'default')

    def test_get_dependency_include_dirs(self, python_class):
        r"""Test get_dependency_include_dirs."""
        with pytest.raises(ValueError):
            python_class.get_dependency_include_dirs('invalid')
        assert(python_class.get_dependency_include_dirs(__file__)
               == [os.path.dirname(__file__)])
        assert(python_class.get_dependency_include_dirs(
            'invalid', default='default') == ['default'])

    def test_get_dependency_order(self, python_class):
        r"""Test get_dependency_order."""
        deps = list(python_class.internal_libraries.keys())
        python_class.get_dependency_order(deps)

    def test_get_flags(self, python_class):
        r"""Test get_flags."""
        compiler = python_class.get_tool('compiler')
        if compiler:
            if ((compiler.combine_with_linker
                 or compiler.no_separate_linking)):
                print(compiler, compiler.get_flags(invalid_kw=True,
                                                   unused_kwargs={},
                                                   libraries=[]))
            else:
                with pytest.raises(ValueError):
                    compiler.get_flags()

    def test_get_linker_flags(self, python_class):
        r"""Test get_linker_flags."""
        if python_class.get_tool('archiver') is False:
            with pytest.raises(RuntimeError):
                python_class.get_linker_flags(libtype='static')
        else:
            python_class.get_linker_flags(libtype='static', for_model=True,
                                          use_library_path_internal=True)
        if getattr(python_class.get_tool('linker'), 'is_dummy', False):
            with pytest.raises(RuntimeError):
                python_class.get_linker_flags(libtype='shared')
        else:
            python_class.get_linker_flags(libtype='shared', for_model=True,
                                          use_library_path=True)
            python_class.get_linker_flags(libtype='shared', for_model=True,
                                          skip_library_libs=True,
                                          use_library_path=True)
            python_class.get_linker_flags(libtype='shared', for_model=True,
                                          skip_library_libs=True,
                                          use_library_path_internal=True)

    def test_executable_command(self, python_class):
        r"""Test executable_command."""
        with pytest.raises(ValueError):
            python_class.executable_command(['test'], exec_type='invalid')
        if python_class.get_tool('compiler').no_separate_linking:
            with pytest.raises(RuntimeError):
                python_class.executable_command(['test'], exec_type='linker')
        else:
            python_class.executable_command(['test'], exec_type='linker')

    def test_compiler_call(self, python_class):
        r"""Test compiler call."""
        tool = python_class.get_tool('compiler')
        assert(tool.call('args', out='test',
                         dry_run=True, skip_flags=True) == '')
        src = [x + tool.get_language_ext()[0] for x in ['args1', 'args2']]
        with pytest.raises(ValueError):
            tool.call(src, out='out1', dont_link=True)
        kwargs = dict(dry_run=True, working_dir=os.getcwd())
        if python_class.language in ['cmake']:
            src = src[:1]
            kwargs['target'] = tool.file2base(src[0])
        tool.call(src, dont_link=True, **kwargs)
        tool.call(src, **kwargs)

    def test_configure(self, python_class):
        r"""Test configuration (presumably after it has already been done)."""
        python_class.configure(ygg_cfg)
        
    def test_get_output_file(self, python_class, source):
        r"""Test get_output_file."""
        tool = python_class.get_tool('compiler')
        fname = source[0]
        tool.get_output_file([fname])

    def test_invalid_function_param2(self, python_class, instance_kwargs):
        r"""Test errors raise during class creation when parameters are invalid."""
        kwargs = copy.deepcopy(instance_kwargs)
        kwargs['name'] = 'test'
        kwargs['args'] = ['test']
        kwargs['function'] = 'invalid'
        kwargs['source_files'] = []
        # With source_files
        if python_class.language_ext:
            kwargs['source_files'] = ['invalid' + python_class.language_ext[0]]
        with pytest.raises(ValueError):
            python_class(**kwargs)
        
    def test_compilers(self, python_class, instance):
        r"""Test available compilers."""
        # Record old tools
        old_tools = {}
        for k in ['compiler', 'linker', 'achiver']:
            old_tools['%s_tool' % k] = getattr(instance,
                                               '%s_tool' % k, None)
        for k in ['compiler_flags', 'linker_flags']:
            old_tools[k] = getattr(instance, k, None)
            setattr(instance, k, [])
        # Compile with each compiler
        for k, v in python_class.get_available_tools('compiler').items():
            if not v.is_installed():
                continue  # pragma: debug
            setattr(instance, 'compiler_tool', v)
            setattr(instance, 'linker_tool', v.linker())
            setattr(instance, 'archiver_tool', v.archiver())
            instance.compile_model(use_ccache=True)
        # Restore the old tools
        for k, v in old_tools.items():
            setattr(instance, k, v)

    def test_compile_model(self, instance, source):
        r"""Test compile model with alternate set of input arguments."""
        fname = source[0]
        with pytest.raises(RuntimeError):
            instance.compile_model(out=os.path.basename(fname),
                                   working_dir=os.path.dirname(fname),
                                   overwrite=True)
        if not instance.is_build_tool:
            instance.compile_model(out=instance.model_file,
                                   overwrite=False)
            assert(os.path.isfile(instance.model_file))
            instance.compile_model(out=instance.model_file,
                                   overwrite=False)
            assert(os.path.isfile(instance.model_file))
            os.remove(instance.model_file)

    def test_call_linker(self, instance):
        r"""Test call_linker with static."""
        out = instance.compile_model(dont_link=True, out=None)
        instance.call_linker(out, for_model=True,
                             working_dir=instance.working_dir,
                             libtype='static')

    def test_parse_arguments(self, python_class, instance):
        r"""Run test to initialize driver using the executable."""
        x = os.path.splitext(instance.source_files[0])[0] + '.out'
        new_inst = python_class('test_name', [x], skip_compile=True)
        assert(new_inst.model_file == x)
        assert(new_inst.source_files == instance.source_files[:1])



import os
from yggdrasil.languages.Python.YggInterface import YggRpcClient


def model_function(in_buf):
    # Get the number of threads from an environment variable set in the yaml
    nthreads = int(os.environ["NTHREAD"])
    for i in range(nthreads):
    
        # The global_scope keyword is required to ensure that the comm
        # persists between function calls
        rpc = YggRpcClient('server_client', global_scope=True)
        print("client(Python:%d): %s" % (i, in_buf))
        ret, result = rpc.call(in_buf)
        if not ret:
            raise RuntimeError('client(Python:%d): RPC CALL ERROR' % i)
        out_buf = result

    return out_buf



import pytest
import os
from yggdrasil import platform
from yggdrasil.drivers.CModelDriver import CModelDriver, LDLinker


def test_LDLinker_tool_version():
    r"""Test the tool_version method of the LDLinker class."""
    if LDLinker.is_installed():
        LDLinker.tool_version()
        LDLinker.get_search_path()
        LDLinker.get_env_flags()


@pytest.mark.skipif(not platform._is_linux, reason="OS is not Linux")
def test_update_ld_library_path():
    r"""Test update_ld_library_path method."""
    lang_dir = CModelDriver.get_language_dir()
    total = os.pathsep.join(['test', lang_dir])
    env = {'LD_LIBRARY_PATH': 'test'}
    env = CModelDriver.update_ld_library_path(env)
    assert(env['LD_LIBRARY_PATH'] == total)
    # Second time to ensure that path not added twice
    env = CModelDriver.update_ld_library_path(env)
    assert(env['LD_LIBRARY_PATH'] == total)


def test_GCCCompiler_dll2a():
    r"""Test the dll2a method of the GCCCompiler class."""
    gcc = CModelDriver.get_tool('compiler', toolname='gcc',
                                default=None)
    if gcc:
        kws = {'toolname': 'gcc'}
        if platform._is_win:
            kws['libtype'] = 'shared'
        dll = CModelDriver.get_dependency_library('python', **kws)
        gcc.dll2a(dll, overwrite=True)
    else:
        with pytest.raises(NotImplementedError):
            CModelDriver.get_tool('compiler', toolname='gcc')



import argparse
import numpy as np


def calculate_concentration(doy, dist, height, offset=60.0):
    r"""Function that calculates the concentration of CO2.

    Args:
        doy (float): Day of year.
        dist (float): Distance from the plant in cm.
        height (float): Distance from the ground in cm.
        offset (float, optional): Offset in the year in days. Defaults to 60.

    Returns:
        float: CO2 concentration in cm^-3

    """
    return np.sin(2.0 * np.pi * (doy + offset) / 365) / (dist * dist * height)


# Parse command-line arguments
parser = argparse.ArgumentParser("Calculate the co2 concentration at a given distance from a plant.")
parser.add_argument('dist', help='Distance from the plant (cm)', type=float)
parser.add_argument('height', help='Distance from the ground (cm)', type=float)
parser.add_argument('doy', help='Day of year', type=float)
args = parser.parse_args()
dist = args.dist
height = args.height
doy = args.doy
offset = 60.0

# Compute concentration
conc = calculate_concentration(doy, dist, height, offset=offset)
print('Concentration', conc)



def get_testing_options(example_name='timesync1'):
    r"""Get testing parameters for this example."""
    def validation_function():
        import os
        import tempfile
        from yggdrasil.examples.timesync1.plot_timesync import main
        tempdir = tempfile.gettempdir()
        fileA = os.path.join(tempdir, "modelA_output.txt")
        fileB = os.path.join(tempdir, "modelB_output.txt")
        main(fileA, fileB, example_name)
    return dict(
        validation_function=validation_function,
        env={'TIMESYNC_TSTEP_A': '20', 'TIMESYNC_TSTEP_B': '3'})



import pytest
from tests.metaschema.properties.test_MetaschemaProperty import (
    TestMetaschemaProperty as base_class)


class TestClassMetaschemaProperty(base_class):
    r"""Test class for ClassMetaschemaProperty class."""
    
    _mod = 'yggdrasil.metaschema.properties.ClassMetaschemaProperty'
    _cls = 'ClassMetaschemaProperty'
    
    @pytest.fixture(scope="class")
    def valid(self):
        r"""Objects that are valid."""
        return [(int(1), int), (dict(), dict)]

    @pytest.fixture(scope="class")
    def invalid(self):
        r"""Objects that are invalid."""
        return [(int(1), dict), (dict(), int)]

    @pytest.fixture(scope="class")
    def valid_compare(self):
        r"""Objects that successfully compare."""
        return [(int, int), (int, (int, float)),
                ([int, float], int),
                ([int, float], (dict, float))]

    @pytest.fixture(scope="class")
    def invalid_compare(self):
        r"""Objects that do not successfully compare."""
        return [(int, float), (int, (dict, float)),
                ((int, dict), float),
                ((int, float), (dict, list))]



import pytest
from tests import TestComponentBase as base_class
import collections
import numpy as np
import copy
from yggdrasil.communication import new_comm


@pytest.mark.usefixtures("pandas_equality_patch")
class TestTransformBase(base_class):
    r"""Test for TransformBase communication flass."""

    _component_type = 'transform'
    
    @pytest.fixture(scope="class", autouse=True)
    def component_subtype(self, transform):
        r"""Subtype of component being tested."""
        return transform

    @pytest.fixture(scope="class", autouse=True)
    def transform(self, request):
        r"""str: Transformation being tested."""
        return request.param

    @pytest.fixture
    def testing_options(self, python_class, options):
        r"""Testing options."""
        if 'explicit_testing_options' in options:
            return copy.deepcopy(options['explicit_testing_options'])
        return python_class.get_testing_options(**options)
    
    @pytest.fixture
    def instance_kwargs(self, testing_options):
        r"""Keyword arguments for a new instance of the tested class."""
        out = {}
        if testing_options:
            out = dict(testing_options[0].get('kwargs', {}))
        return out

    @pytest.fixture(scope="class")
    def assert_iterator_equality(self, nested_approx):
        r"""Check for equality between iterators."""
        from collections.abc import Iterator

        def iterator_equality_w(a, b):
            if isinstance(a, Iterator) and isinstance(b, Iterator):
                assert(list(copy.deepcopy(a))
                       == nested_approx(list(copy.deepcopy(b))))
            else:
                assert(a == nested_approx(b))
        return iterator_equality_w

    def test_transform(self, python_class, testing_options,
                       assert_iterator_equality):
        r"""Test transform."""
        for x in testing_options:
            inst = python_class(**x.get('kwargs', {}))
            for msg_in, msg_exp in x.get('in/out', []):
                # if isinstance(msg_in, collections.abc.Iterator):
                #     msg_in0 = copy.deepcopy(msg_in)
                # else:
                #     msg_in0 = msg_in
                if isinstance(msg_exp, type(BaseException)):
                    with pytest.raises(msg_exp):
                        inst(msg_in)
                else:
                    msg_out = inst(msg_in)
                    assert_iterator_equality(msg_out, msg_exp)

    def test_transform_type(self, python_class, testing_options):
        r"""Test transform_type."""
        for x in testing_options:
            inst = python_class(**x.get('kwargs', {}))
            for typ_in, typ_exp in x.get('in/out_t', []):
                if isinstance(typ_exp, type(BaseException)):
                    with pytest.raises(typ_exp):
                        inst.validate_datatype(typ_in)
                else:
                    inst.validate_datatype(typ_in)
                    typ_out = inst.transform_datatype(typ_in)
                    assert(typ_out == typ_exp)

    def test_transform_empty(self, instance):
        r"""Test transform of empty bytes message."""
        assert(instance(b'', no_init=True) == b'')

    def test_send_comm(self, class_name, python_class, testing_options,
                       timeout, nested_approx):
        r"""Test transform within send comm."""
        if class_name in ['PandasTransform', 'ArrayTransform']:
            pytest.skip("Transformation negated on send")
        for x in testing_options:
            if (((not x.get('in/out', False))
                 or isinstance(x['in/out'][0][1],
                               type(BaseException)))):  # pragma: debug
                continue
            msg_in, msg_out = x['in/out'][0]
            send_comm = new_comm('test_send', reverse_names=True,
                                 direction='send', use_async=False,
                                 transform=[python_class(**x.get('kwargs', {}))])
            recv_comm = new_comm('test_recv', **send_comm.opp_comm_kwargs())
            if isinstance(msg_in, collections.abc.Iterator):
                msg_out_list = list(msg_out)
                msg_out = iter(msg_out_list)
            else:
                msg_out_list = [msg_out]
            try:
                for imsg_out in msg_out_list:
                    flag = send_comm.send(msg_in, timeout=timeout)
                    assert(flag)
                    if class_name in ['IterateTransform']:
                        for iimsg_out in imsg_out:
                            flag, msg_recv = recv_comm.recv(timeout=timeout)
                            assert(flag)
                            assert(msg_recv == nested_approx(iimsg_out))
                        msg_recv = imsg_out
                    elif ((class_name in ['FilterTransform'])
                          and isinstance(imsg_out, collections.abc.Iterator)):
                        assert(recv_comm.n_msg_recv == 0)
                        flag, msg_recv = recv_comm.recv(timeout=0.0)
                        assert(flag)
                        assert(msg_recv == b'')
                        msg_recv = imsg_out
                    elif ((class_name in ['MapFieldsTransform',
                                          'SelectFieldsTransform'])
                          and isinstance(imsg_out, np.ndarray)):
                        flag, msg_recv = recv_comm.recv_array(timeout=timeout)
                    else:
                        flag, msg_recv = recv_comm.recv(timeout=timeout)
                    assert(flag)
                    assert(msg_recv == nested_approx(imsg_out))
            finally:
                send_comm.close(linger=True)
                recv_comm.close(linger=True)
                send_comm.disconnect()
                recv_comm.disconnect()
                del send_comm
                del recv_comm

    def test_recv_comm(self, class_name, python_class, testing_options,
                       timeout, assert_iterator_equality):
        r"""Test transform within recv comm."""
        for x in testing_options:
            if (((not x.get('in/out', False))
                 or isinstance(x['in/out'][0][1],
                               type(BaseException)))):  # pragma: debug
                continue
            msg_in, msg_out = x['in/out'][0]
            send_comm = new_comm('test_send', reverse_names=True,
                                 direction='send', use_async=False)
            recv_comm = new_comm('test_recv',
                                 transform=[python_class(**x.get('kwargs', {}))],
                                 **send_comm.opp_comm_kwargs())
            if isinstance(msg_in, collections.abc.Iterator):
                msg_out_list = list(msg_out)
                msg_out = iter(msg_out_list)
            else:
                msg_out_list = [msg_out]
            try:
                flag = send_comm.send(msg_in, timeout=timeout)
                assert(flag)
                for imsg_out in msg_out_list:
                    if (((class_name in ['MapFieldsTransform',
                                         'SelectFieldsTransform'])
                         and isinstance(imsg_out, np.ndarray))):
                        flag, msg_recv = recv_comm.recv_array(timeout=timeout)
                    else:
                        flag, msg_recv = recv_comm.recv(timeout=timeout)
                    assert(flag)
                    assert_iterator_equality(msg_recv, imsg_out)
            finally:
                send_comm.close(linger=True)
                recv_comm.close(linger=True)
                send_comm.disconnect()
                recv_comm.disconnect()
                del send_comm
                del recv_comm



import os
import copy
import pytest
import itertools
import shutil
from yggdrasil.services import (
    IntegrationServiceManager, ServerError, validate_model_submission)
from yggdrasil.examples import yamls as ex_yamls
from yggdrasil import runner, import_as_function


def _make_ids(ids):
    return ','.join([str(x) for x in ids])


@pytest.mark.language('c')
@pytest.mark.language('c++')
def test_call_integration_remote():
    r"""Test with remote integration service."""
    name = 'photosynthesis'
    test_yml = ex_yamls['fakeplant']['python']
    copy_yml = ex_yamls['fakeplant']['c'][0]
    remote_yml = '_remote'.join(os.path.splitext(test_yml))
    yamls = copy.copy(ex_yamls['fakeplant']['all_nomatlab'])
    yamls.remove(test_yml)
    yamls.remove(copy_yml)
    yamls.append(remote_yml)
    address = 'https://model-service-demo.herokuapp.com/'
    service_type = 'flask'
    cli = IntegrationServiceManager(service_type=service_type,
                                    for_request=True,
                                    address=address)
    cli.wait_for_server()
    if not cli.is_running:  # pragma: debug
        pytest.skip("Heroku app is not running.")
    try:
        shutil.copy(copy_yml, remote_yml)
        with open(remote_yml, 'a') as fd:
            fd.write('\n'.join(['service:',
                                f'    name: {name}',
                                f'    type: {service_type}',
                                f'    address: {address}']))
        r = runner.get_runner(yamls)
        r.run()
        assert(not r.error_flag)
    finally:
        if os.path.isfile(remote_yml):
            os.remove(remote_yml)
        cli.send_request(name, action='stop')
        
        
class TestServices(object):
    r"""Class to test integration services."""

    @pytest.fixture(params=itertools.product(['flask', 'rmq'], [None, 'rmq']),
                    ids=_make_ids, scope="class", autouse=True)
    def running_service(self, request, running_service):
        with running_service(request.param[0], request.param[1]) as cli:
            self.cli = cli
            yield cli
            self.cli = None

    def call_integration_service(self, cli, yamls, test_yml, copy_yml=None,
                                 name='test', yaml_param=None):
        r"""Call an integration that includes a service."""
        remote_yml = '_remote'.join(os.path.splitext(test_yml))
        yamls = copy.copy(yamls)
        yamls.remove(test_yml)
        if copy_yml:
            yamls.remove(copy_yml)
        yamls.append(remote_yml)
        service_type = cli.service_type
        try:
            address = cli.address
            if copy_yml:
                shutil.copy(copy_yml, remote_yml)
                remote_code = 'a'
            else:
                remote_code = 'w'
            lines = ['service:',
                     f'    name: {name}',
                     f'    yamls: [{test_yml}]',
                     f'    type: {service_type}',
                     f'    address: {address}']
            if yaml_param:
                lines.append('    yaml_param:')
                for k, v in yaml_param.items():
                    lines.append(f'        {k}: "{v}"')
            with open(remote_yml, remote_code) as fd:
                fd.write('\n'.join(lines))
            r = runner.get_runner(yamls)
            r.run()
            assert(not r.error_flag)
        finally:
            if os.path.isfile(remote_yml):
                os.remove(remote_yml)

    def test_git_fails(self, running_service):
        r"""Test that sending a request for a git YAML fails."""
        cli = running_service
        test_yml = ("git:https://github.com/cropsinsilico/example-fakemodel/"
                    "fakemodel3.yml")
        assert(not os.path.isfile(
            "cropsinsilico/example-fakemodel/fakemodel3.yml"))
        with pytest.raises(ServerError):
            cli.send_request(yamls=test_yml, action='start')

    def test_integration_service(self, running_service):
        r"""Test starting/stopping an integration service via flask/rmq."""
        cli = running_service
        test_yml = ex_yamls['fakeplant']['python']
        with pytest.raises(ServerError):
            cli.send_request(test_yml, action='invalid')
        print(cli.send_request(test_yml))
        cli.printStatus()
        if cli.service_type == 'flask':
            import requests
            r = requests.get(cli.address)
            r.raise_for_status()
        cli.send_request(test_yml, action='status')
        cli.send_request(action='status', client_id=None)
        cli.send_request(test_yml, yamls=test_yml, action='stop')
        with pytest.raises(ServerError):
            cli.send_request(['invalid'], action='stop')
        cli.printStatus(return_str=True)
        cli.send_request([test_yml])
        cli.send_request([test_yml], action='stop')

    def test_registered_service(self, running_service):
        r"""Test registering an integration service."""
        if (((running_service.commtype != 'rest')
             or (running_service.service_type != 'flask'))):
            pytest.skip("redundent test")
        cli = running_service
        test_yml = ex_yamls['fakeplant']['python']
        with pytest.raises(KeyError):
            cli.registry.remove('test')
        with pytest.raises(ServerError):
            cli.send_request('test')
        print(cli.send_request('FakePlant'))
        cli.registry.add('test', test_yml, namespace='remote')
        print(cli.send_request('test'))
        with pytest.raises(ValueError):
            cli.registry.add('test', [test_yml])
        cli.send_request('test', action='stop')
        # cli.stop_server()
        cli.registry.remove('test')
        with pytest.raises(KeyError):
            cli.registry.remove('test')
        # Register from file
        reg_coll = os.path.join(os.path.dirname(test_yml),
                                'registry_collection.yml')
        test_yml_base = os.path.basename(test_yml)
        with open(reg_coll, 'w') as fd:
            fd.write(f'photosynthesis:\n  - {test_yml_base}')
        try:
            cli.registry.add(reg_coll)
            print(cli.send_request('photosynthesis', namespace='phot'))
            with pytest.raises(ValueError):
                cli.registry.add('photosynthesis', [test_yml], invalid=1)
            cli.send_request('photosynthesis', action='stop')
            cli.registry.remove(reg_coll)
            with pytest.raises(KeyError):
                cli.registry.remove('photosynthesis')
        finally:
            os.remove(reg_coll)

    @pytest.mark.language('c')
    @pytest.mark.language('c++')
    def test_calling_integration_service(self, running_service):
        r"""Test calling an integrations as a service in an integration."""
        self.call_integration_service(
            running_service,
            ex_yamls['fakeplant']['all_nomatlab'],
            ex_yamls['fakeplant']['python'],
            copy_yml=ex_yamls['fakeplant']['c'][0])

    @pytest.mark.language('c')
    @pytest.mark.language('c++')
    def test_calling_server_as_service(self, running_service):
        r"""Test calling an integration service that is a server in an
        integration."""
        if (((running_service.commtype != 'rest')
             or (running_service.service_type != 'flask'))):
            pytest.skip("redundent test")  # pragma: testing
        yaml_param = dict(FIB_ITERATIONS='3',
                          FIB_SERVER_SLEEP_SECONDS='0.01')
        os.environ.update(yaml_param)
        yamls = ex_yamls['rpcFib']['all_nomatlab']
        service = None
        for x in yamls:
            if 'Srv' in x:
                service = x
                break
        self.call_integration_service(running_service, yamls, service,
                                      name='rpcFibSrv',
                                      yaml_param=yaml_param)

    def test_calling_service_as_function(self, running_service):
        r"""Test calling an integrations as a service in an integration."""
        # if running_service.commtype != None:
        #     pytest.skip("redundent test")
        cli = running_service
        name = 'test'
        test_yml = ex_yamls['fakeplant']['python']
        try:
            cli.registry.add(name, test_yml)
            fmodel = import_as_function(name, cli.address)
            input_args = {}
            for x in fmodel.arguments:
                input_args[x] = 1.0
            fmodel.model_info()
            result = fmodel(**input_args)
            for x in fmodel.returns:
                assert(x in result)
            result = fmodel(*list(input_args.values()))
            for x in fmodel.returns:
                assert(x in result)
            fmodel.stop()
            fmodel.stop()
        finally:
            cli.registry.remove(name)


def test_validate_model_submission():
    r"""Test validate_model_submission"""
    import git
    try:
        fname = os.path.join(os.path.dirname(__file__), 'yamls',
                             'FakePlant.yaml')
        validate_model_submission([fname])
        os.remove(os.path.join('cropsinsilico', 'example-fakemodel',
                               'LICENSE'))
        with pytest.raises(RuntimeError):
            validate_model_submission(fname)
    finally:
        if os.path.isfile('cropsinsilico/example-fakemodel/fakemodel.yml'):
            git.rmtree("cropsinsilico")


def test_validate_model_repo():
    r"""Test validation of YAMLs in the model repository."""
    import git
    import tempfile
    dest = os.path.join(tempfile.gettempdir(), "model_repo")
    url = "https://github.com/cropsinsilico/yggdrasil_models"
    for x in [url, url + "_test"]:
        try:
            repo = git.Repo.clone_from(x, dest)
            model_dir = os.path.join(dest, "models")
            if os.path.isdir(model_dir):
                # This condition can be removed once there are models in the
                # non-dev repository
                validate_model_submission(model_dir)
            repo.close()
        finally:
            if os.path.isdir(dest):
                git.rmtree(dest)



import pytest
import os
import tempfile
from yggdrasil import tools, platform
from tests import TestClassBase as base_class


class DummyTarget(object):
    r"""Class to wrap for testing the ProxyObject class."""

    def __init__(self):
        self.b = 2

    def __call__(self, x):
        print(x)

    def __bytes__(self):
        return b'hello'

    def __hash__(self):
        return 1

    def __eq__(self, other):
        return True


class DummyProxyObject(tools.ProxyObject):
    r"""Class for testing corner cases of the ProxyObject."""

    __slots__ = ['a']

    def __init__(self, *args, **kwargs):
        self.a = 1
        super(DummyProxyObject, self).__init__(*args, **kwargs)


def test_ProxyObject():
    r"""Test corner casses of the ProxyObject class."""
    x = DummyProxyObject(DummyTarget())
    repr(x)
    str(x)
    bytes(x)
    bool(x)
    hash(x)
    x(1)
    assert(x == 1)
    for k in ['a', 'b']:
        delattr(x, k)
        assert(not hasattr(x, k))


def make_temp(fname_base, count=1):
    r"""Create temporary copies of same file with different extensions."""
    fname_base = fname_base.lower()
    tempdir = os.path.normcase(os.path.normpath(tempfile.gettempdir()))
    if (tempdir + os.pathsep) not in os.environ['PATH']:
        os.environ['PATH'] = os.pathsep.join([tempdir, os.environ.get('PATH')])
    fname_pattern = fname_base + '.*'
    fname = os.path.join(tempdir, fname_base)
    out = []
    for i in range(count):
        fname_i = '%s.%d' % (fname, i)
        out.append(fname_i)
        if not os.path.isfile(fname_i):
            with open(fname_i, 'w') as fd:
                fd.write('Test file %d' % i)
    return tempdir, fname_pattern, out


def make_temp_single():
    r"""Create single temporary file."""
    return make_temp('single_test_file')


def make_temp_multiple():
    r"""Create multiple temporary files."""
    return make_temp('multiple_test_file', count=2)


def test_bytes2str():
    r"""Test bytes2str."""
    vals = [(b'hello', 'hello'),
            ('hello', 'hello'),
            ((b'a', b'b'), ('a', 'b')),
            ({'a': b'a', 'b': b'b'}, {'a': 'a', 'b': 'b'}),
            ([b'a', b'b'], ['a', 'b']),
            ([b'a', [b'b', b'c']], ['a', ['b', 'c']])]
    for x, exp in vals:
        assert(tools.bytes2str(x, recurse=True) == exp)


def test_str2bytes():
    r"""Test str2bytes."""
    vals = [('hello', b'hello'),
            (b'hello', b'hello'),
            (('a', 'b'), (b'a', b'b')),
            ({'a': 'a', 'b': 'b'}, {'a': b'a', 'b': b'b'}),
            (['a', 'b'], [b'a', b'b']),
            (['a', ['b', 'c']], [b'a', [b'b', b'c']])]
    for x, exp in vals:
        assert(tools.str2bytes(x, recurse=True) == exp)


def test_timer_context():
    r"""Test timer_context."""
    with tools.timer_context("Test timeout: {elapsed}"):
        print("timer context body")


def test_display_source():
    r"""Test display_source."""
    fname = os.path.abspath(__file__)
    tools.display_source([fname], number_lines=True)
    tools.display_source([fname], return_lines=True)
    tools.display_source(test_display_source)
    fname_txt = os.path.join(tempfile.gettempdir(), 'unknown_example.invalid')
    with open(fname_txt, 'w') as fd:
        fd.write('hello')
    try:
        tools.display_source(fname_txt)
    finally:
        os.remove(fname_txt)


def test_display_source_diff():
    r"""Test display_source_diff."""
    fname1 = os.path.abspath(__file__)
    fname2 = os.path.join(tempfile.gettempdir(), os.path.basename(fname1))
    with open(fname1, 'r') as fd:
        lines = fd.read()
    with open(fname2, 'w') as fd:
        fd.write(lines[:int(len(lines) / 2)] + 'additional')
    try:
        tools.display_source_diff(fname1, fname2, number_lines=True)
        tools.display_source_diff(fname1, fname2, return_lines=True)
        tools.display_source_diff(test_display_source_diff,
                                  test_display_source_diff)
    finally:
        os.remove(fname2)


def test_get_shell():
    r"""Test get_shell."""
    tools.get_shell()


def test_in_powershell():
    r"""Test in_powershell."""
    tools.in_powershell()


def test_get_conda_prefix():
    r"""Test get_conda_prefix."""
    tools.get_conda_prefix()

    
def test_get_conda_env():
    r"""Test get_conda_env."""
    tools.get_conda_env()


def test_get_python_c_library():
    r"""Test get_python_c_library."""
    tools.get_python_c_library(allow_failure=True)


def test_is_language_alias():
    r"""Test is_language_alias."""
    assert(tools.is_language_alias('c++', 'cpp'))
    assert(tools.is_language_alias('r', 'R'))
    assert(tools.is_language_alias('MATLAB', 'matlab'))
    

def test_get_supported():
    r"""Test get_supported_<platforms/lang/comm>."""
    tools.get_supported_platforms()
    tools.get_supported_lang()
    tools.get_supported_comm()


def test_get_installed():
    r"""Test get_installed_<lang/comm>."""
    tools.get_installed_lang()
    tools.get_installed_comm()


def test_is_comm_installed():
    r"""Test is_comm_installed for any."""
    assert(tools.is_comm_installed('zmq', language='any'))


def test_locate_file():
    r"""Test file location method."""
    # Missing file
    assert(not tools.locate_file('missing_file.fake'))
    assert(not tools.locate_file(['missing_file.fake']))
    # Single file
    sdir, spat, sans = make_temp_single()
    sout = tools.locate_file(spat, verification_func=os.path.isfile)
    assert(isinstance(sout, (bytes, str)))
    assert(sout == sans[0])
    # Multiple files
    mdir, mpat, mans = make_temp_multiple()
    with pytest.warns(RuntimeWarning):
        mout = tools.locate_file([mpat], show_alternates=True)
        assert(isinstance(mout, (bytes, str)))
        assert(mout == mans[0])
    

def test_find_all():
    r"""Test find_all."""
    # Missing file
    assert(not tools.find_all('missing_file.fake', 'invalid'))
    # Single file
    sdir, spat, sans = make_temp_single()
    sout = tools.find_all(spat, sdir)
    assert(isinstance(sout, list))
    assert(sout == sans)
    # Multiple files
    mdir, mpat, mans = make_temp_multiple()
    mout = tools.find_all(mpat, mdir)
    assert(isinstance(mout, list))
    assert(mout == mans)


def test_locate_path():
    r"""Test file location."""
    # Search for current file
    fdir, fname = os.path.split(__file__)
    basedir = os.path.dirname(fdir)
    fpath = tools.locate_path(fname, basedir=basedir)
    assert(fpath)
    assert(__file__ in fpath)
    # assert(__file__ == fpath)
    # Search for invalid file
    fname = 'invalid_file.ext'
    fpath = tools.locate_path(fname, basedir=basedir)
    assert(not fpath)


def test_popen_nobuffer():
    r"""Test open of process without buffer."""
    ans = os.getcwd()  # + '\n'
    # Test w/o shell
    if platform._is_win:  # pragma: windows
        args = ['cmd', '/c', 'cd']
    else:
        args = ['pwd']
    p = tools.popen_nobuffer(args)
    out, err = p.communicate()
    res = tools.bytes2str(out).splitlines()[0]
    assert(res == ans)
    # Test w/ shell
    if platform._is_win:  # pragma: windows
        args = 'cd'
    else:
        args = 'pwd'
    p = tools.popen_nobuffer(args, shell=True)
    out, err = p.communicate()
    res = tools.bytes2str(out).splitlines()[0]
    assert(res == ans)


def test_eval_kwarg():
    r"""Ensure strings & objects properly evaluated."""
    vals = [None, True, False, ['one', 'two'], 'one']
    for v in vals:
        assert(tools.eval_kwarg(v) == v)
        assert(tools.eval_kwarg(str(v)) == v)
    assert(tools.eval_kwarg("'one'") == 'one')
    assert(tools.eval_kwarg('"one"') == 'one')


class TestYggClass(base_class):
    r"""Test basic behavior of YggTestClass."""

    _cls = 'YggClass'
    _mod = 'yggdrasil.tools'

    @pytest.fixture
    def namespace(self, uuid):
        return f'TESTING_{uuid}'

    @pytest.fixture
    def instance_kwargs(self, timeout, polling_interval):
        r"""Keyword arguments for a new instance of the tested class."""
        return {'timeout': timeout,
                'sleeptime': polling_interval}
        
    def test_prints(self, instance):
        r"""Test logging at various levels."""
        instance.display(1)
        instance.info(1)
        instance.debug(1)
        instance.verbose_debug(1)
        instance.critical(1)
        instance.warning(1)
        instance.warn(1)
        instance.error(1)
        instance.exception(1)
        try:
            raise Exception("Test exception")
        except Exception:
            instance.exception(1)
        instance.printStatus()
        instance.printStatus(return_str=True)
        instance.special_debug(1)
        instance.suppress_special_debug = True
        instance.special_debug(1)
        instance.suppress_special_debug = False
        instance.interface_info(1)
        instance.language_info([])(1)
        instance.language_info('python')(1)

    def test_wait_on_function(self, instance):
        r"""Test functionality of async wait on function."""
        def func():
            return False
        instance.wait_on_function(func, timeout=0.0,
                                  polling_interval=0.0,
                                  quiet=True)

    def test_timeout(self, instance):
        r"""Test functionality of timeout."""
        # Test w/o timeout
        instance.start_timeout(10, key='fake_key')
        assert(not instance.check_timeout(key='fake_key'))
        # Test errors
        with pytest.raises(KeyError):
            instance.start_timeout(0.1, key='fake_key')
        instance.stop_timeout(key='fake_key')
        with pytest.raises(KeyError):
            instance.check_timeout()
        with pytest.raises(KeyError):
            instance.check_timeout(key='fake_key')
        with pytest.raises(KeyError):
            instance.stop_timeout(key='fake_key')
        # Test w/ timeout
        T = instance.start_timeout(0.001)  # instance.sleeptime)
        while not T.is_out:
            instance.sleep()
        assert(instance.check_timeout())
        instance.stop_timeout(quiet=True)



import pytest
from tests.metaschema.datatypes.test_MetaschemaType import (
    TestMetaschemaType as base_class)
import numpy as np


class TestJSONBooleanMetaschemaType(base_class):
    r"""Test class for JSONBooleanMetaschemaType class."""
    
    _mod = 'yggdrasil.metaschema.datatypes.JSONMetaschemaType'
    _cls = 'JSONBooleanMetaschemaType'

    @pytest.fixture(scope="class")
    def valid_encoded(self, python_class):
        r"""list: Encoded objects that are valid under this type."""
        return [{'type': python_class.name}]

    @pytest.fixture(scope="class")
    def valid_decoded(self):
        r"""list: Objects that are valid under this type."""
        return [True, False]
    
    @pytest.fixture(scope="class")
    def valid_normalize(self):
        r"""list: Pairs of pre-/post-normalized objects."""
        return [('True', True), ('true', True),
                ('False', False), ('false', False),
                ('hello', 'hello')]


class TestJSONIntegerMetaschemaType(TestJSONBooleanMetaschemaType):
    r"""Test class for JSONIntegerMetaschemaType class."""
    
    _cls = 'JSONIntegerMetaschemaType'

    @pytest.fixture(scope="class")
    def valid_decoded(self):
        r"""list: Objects that are valid under this type."""
        return [int(1), np.int(1)]
    
    @pytest.fixture(scope="class")
    def valid_normalize(self):
        r"""list: Pairs of pre-/post-normalized objects."""
        return [('1', 1), ('hello', 'hello')]


class TestJSONNullMetaschemaType(TestJSONBooleanMetaschemaType):
    r"""Test class for JSONNullMetaschemaType class."""
    
    _cls = 'JSONNullMetaschemaType'

    @pytest.fixture(scope="class")
    def valid_decoded(self):
        r"""list: Objects that are valid under this type."""
        return [None]
    
    @pytest.fixture(scope="class")
    def invalid_validate(self):
        r"""list: Objects that are invalid under this type."""
        return ['hello']

    @pytest.fixture(scope="class")
    def valid_normalize(self):
        r"""list: Pairs of pre-/post-normalized objects."""
        return []


class TestJSONNumberMetaschemaType(TestJSONBooleanMetaschemaType):
    r"""Test class for JSONNumberMetaschemaType class."""
    
    _cls = 'JSONNumberMetaschemaType'

    @pytest.fixture(scope="class")
    def valid_decoded(self):
        r"""list: Objects that are valid under this type."""
        return [int(1), np.int(1), float(1), np.float(1)]
        
    @pytest.fixture(scope="class")
    def valid_normalize(self):
        r"""list: Pairs of pre-/post-normalized objects."""
        return [('1', 1.0), ('1.0', 1.0), ('hello', 'hello')]


class TestJSONStringMetaschemaType(TestJSONBooleanMetaschemaType):
    r"""Test class for JSONStringMetaschemaType class."""
    
    _cls = 'JSONStringMetaschemaType'

    @pytest.fixture(scope="class")
    def valid_decoded(self):
        r"""list: Objects that are valid under this type."""
        return ['hello']
    
    @pytest.fixture(scope="class")
    def valid_normalize(self):
        r"""list: Pairs of pre-/post-normalized objects."""
        return [(1, '1'), (1.0, '1.0'), ([1, 2, 3], [1, 2, 3])]



import os
import numpy as np


temp = np.array([400.0, 410.0, 420.0, 430.0, 440.0, 450.0])  # Temp in K
theta_i = np.linspace(0.0, np.pi / 2, 10)  # radians
n1 = 1.0
n2 = 1.5

# Check if model is running as a part of an yggdrasil integration
with_yggdrasil = os.environ.get('YGG_SUBPROCESS', False)

# If the model is running as part of an yggdrasil integration, import
# the relevant yggdrasil routines and use the interface routine to
# complete the connection defined in the YAML
if with_yggdrasil:
    from yggdrasil import units
    from yggdrasil.languages.Python.YggInterface import YggInput, YggOutput
    temp_in = YggInput('temperature')
    refl_out = YggOutput('reflectance')

    # Receive the temperatures
    flag, temp = temp_in.recv()
    if not flag:
        raise Exception("Error receiving temperature.")

    # Ensure correct units
    temp = units.convert_to(temp[0], 'K')  # list of one table column
    temp = units.get_data(temp)


for i in range(len(temp)):
    in1 = n1 - 0.0001 * (400.0 - temp[i])
    in2 = n2 - 0.05 * (400.0 - temp[i])
    theta_t = np.arcsin(in1 * np.sin(theta_i) / in2)
    R = np.abs((in1 * np.cos(theta_i) - in2 * np.cos(theta_t))
               / (in1 * np.cos(theta_i) + in2 * np.cos(theta_t)))**2
    print("Reflectance at %s K: %s" % (temp[i], R))
    if with_yggdrasil:
        flag = refl_out.send(temp[i], *R)  # expand array elements
        if not flag:
            raise Exception("Error sending reflectance")



import weakref
from yggdrasil.metaschema import MetaschemaTypeError
from yggdrasil.metaschema.properties.MetaschemaProperty import (
    MetaschemaProperty)
from yggdrasil.metaschema.properties.JSONArrayMetaschemaProperties import (
    ItemsMetaschemaProperty)


class ArgsMetaschemaProperty(MetaschemaProperty):
    r"""Property class for 'args' property."""

    name = 'args'
    schema = {'description': ('Arguments required to recreate a class instance.'),
              'type': 'array'}
    _instance_dict_attr = ['input_arguments', 'input_args']

    @classmethod
    def instance2args(cls, instance):
        r"""Get input arguments from a class instance.

        Args:
            instance (object): Instance of a Python class.

        Returns:
            dict: Input arguments for re-creating the instance.

        """
        out = None
        for k in cls._instance_dict_attr:
            if out is not None:
                break
            if hasattr(instance, k):
                out = getattr(instance, k)
            elif hasattr(instance, 'get_' + k):
                out = getattr(instance, 'get_' + k)()
            elif hasattr(instance, '_' + k):
                out = getattr(instance, '_' + k)
        if isinstance(out, (list, tuple)):
            out_real = []
            for x in out:
                if isinstance(x, weakref.ReferenceType):
                    out_real.append(x())
                else:
                    out_real.append(x)
            return out_real
        elif isinstance(out, dict):
            out_real = {}
            for k, v in out.items():
                if isinstance(v, weakref.ReferenceType):
                    out_real[k] = v()
                else:
                    out_real[k] = v
            return out_real
        else:
            raise MetaschemaTypeError('Could not locate dictionary of arguments.')

    @classmethod
    def encode(cls, instance, typedef=None):
        r"""Encoder for the 'args' property."""
        typedef_args = None
        # if isinstance(typedef, dict) and ('args' in typedef):
        #     typedef_args = typedef['args']
        args = cls.instance2args(instance)
        return ItemsMetaschemaProperty.encode(args, typedef_args)

    @classmethod
    def compare(cls, *args, **kwargs):
        r"""Comparison method for 'args' container property."""
        for e in ItemsMetaschemaProperty.compare(*args, **kwargs):
            yield e



import pytest
import pickle
from yggdrasil import multitasking
from tests import TestClassBase as base_class


class TstClass(object):
    
    x = 'test'


class LockedTstClass(multitasking.LockedObject):

    _base_class = TstClass
    _base_attr = ['x']


def test_LockedAttr():
    r"""Test access to locked attribute."""
    z = LockedTstClass()
    z.x
    z.disconnect()
    with pytest.raises(multitasking.AliasDisconnectError):
        z.x


def test_WaitableFunction():
    r"""Test WaitableFunction."""
    def always_false():
        return False
    x = multitasking.WaitableFunction(always_false, polling_interval=0.0)
    assert(not x.wait(timeout=0.0, on_timeout=True))
    with pytest.raises(multitasking.TimeoutError):
        x.wait(timeout=0.0, on_timeout="Error message")


def test_TaskThread():
    r"""Test thread based Task."""
    q = multitasking.Task(target=multitasking.test_target_error)
    assert(not q.is_alive())
    q.start()
    q._base._errored.wait(1.0)
    assert(q._base._errored.is_set())
    q.join(60.0)
    assert(not q.is_alive())


def test_TaskProcess():
    r"""Test process based Task."""
    q = multitasking.Task(target=multitasking.test_target_error,
                          task_method='parallel')
    assert(not q.is_alive())
    q.start()
    q.join(60.0)
    assert(not q.is_alive())


class TestContextThread(base_class):
    r"""Test for thread based Context."""

    _cls = 'Context'
    _mod = 'yggdrasil.multitasking'
    _task_method = 'thread'

    @pytest.fixture(scope="class", autouse=True)
    def reset_test(self, module_name, class_name, task_method):
        self.__class__._first_test = True

    @pytest.fixture(scope="class", autouse=True, params=['thread', 'process'])
    def task_method(self, request):
        r"""Method that should be used for the generatd task."""
        return request.param
    
    @pytest.fixture(scope="class")
    def instance_kwargs(self, task_method):
        r"""Keyword arguments for a new instance of the tested class."""
        return dict(task_method=task_method)
        
    def test_RLock(self, instance):
        r"""Test creation of RLock from context."""
        x = instance.RLock()
        x.disconnect()

    def test_Event(self, instance):
        r"""Test creation of Event from context."""
        x = instance.Event()
        x.disconnect()

    def test_Task(self, instance):
        r"""Test creation of Task from context."""
        x = instance.Task()
        x.disconnect()

    def test_Queue(self, instance):
        r"""Test creation of Queue from context."""
        x = instance.Queue()
        x.disconnect()


class TstContextObject(object):

    _cls = None
    _mod = 'yggdrasil.multitasking'

    @pytest.fixture
    def check_decoded(self):
        r"""Check that object was decoded correctly."""
        def check_decoded_w(decoded):
            pass
        return check_decoded_w

    def test_pickle(self, instance, check_decoded):
        r"""Test pickling and unpickling the object."""
        encoded = pickle.dumps(instance)
        decoded = pickle.loads(encoded)
        check_decoded(decoded)


class TestRLock(TstContextObject, base_class):

    _cls = 'RLock'


class TestEvent(TstContextObject, base_class):

    _cls = 'Event'

    @pytest.fixture
    def check_decoded(self, instance):
        r"""Check that object was decoded correctly."""
        def check_decoded_w(decoded):
            assert(decoded.is_set() == instance.is_set())
        return check_decoded_w

    def test_pickle(self, instance, check_decoded):
        r"""Test pickling and unpickling the object."""
        instance.set()
        super(TestEvent, self).test_pickle(instance, check_decoded)

    def test_callback(self, instance):
        r"""Test callabacks."""
        def set_callback():
            self.state = 'set'

        def clear_callback():
            self.state = 'clear'

        instance.add_callback(set_callback, trigger='set')
        instance.add_callback(clear_callback, trigger='clear')
        instance.set()
        assert(self.state == 'set')
        instance.clear()
        assert(self.state == 'clear')


class TestValueEvent(TestEvent):

    _cls = 'ValueEvent'

    def test_value(self, instance):
        r"""Test setting/clearning event value."""
        assert(instance.get() is None)
        instance.set('test')
        assert(instance.get() == 'test')
        instance.clear()
        assert(instance.get() is None)


class TestTask(TstContextObject, base_class):

    _cls = 'Task'

    @pytest.fixture
    def check_decoded(self, instance):
        r"""Check that object was decoded correctly."""
        def check_decoded_w(decoded):
            assert(decoded.name == instance.name)
            assert(decoded.daemon == instance.daemon)
        return check_decoded_w


class TestQueue(TstContextObject, base_class):

    _cls = 'Queue'

    def test_join(self, instance):
        r"""Test join."""
        instance.join()


class TestYggTask(base_class):
    r"""Test basic behavior of YggTask class."""

    _cls = 'YggTask'
    _mod = 'yggdrasil.multitasking'

    @pytest.fixture(scope="class", autouse=True)
    def reset_test(self, module_name, class_name, task_method):
        self.__class__._first_test = True

    @pytest.fixture(scope="class", autouse=True, params=['thread', 'process'])
    def task_method(self, request):
        r"""Method that should be used for the generatd task."""
        return request.param

    @pytest.fixture(scope="class", autouse=True)
    def context(self, task_method):
        if task_method == 'process':
            return multitasking.mp_ctx_spawn
    
    @pytest.fixture(scope="class")
    def instance_kwargs(self, timeout, polling_interval, task_method,
                        context):
        r"""Keyword arguments for a new instance of the tested class."""
        return dict(timeout=timeout, sleeptime=polling_interval,
                    target=multitasking.test_target_sleep,
                    task_method=task_method,
                    context=context)

    @pytest.fixture
    def namespace(self, uuid):
        r"""Test namespace."""
        return f'TESTING_{uuid}'

    @pytest.fixture
    def instance(self, python_class, instance_args, instance_kwargs):
        r"""New instance of the python class for testing."""
        out = python_class(*instance_args, **instance_kwargs)
        yield out
        # Must disconnect as calling 'del' dosn't trigger garbage collection
        # until after tests due to presence of a ref in pytest fixtures
        out.disconnect()
        
    def test_id(self, instance):
        r"""Test process ID and ident."""
        instance.pid
        instance.ident

    def test_daemon(self, instance):
        r"""Test process/thread daemon property."""
        instance.daemon

    def test_exitcode(self, instance):
        r"""Test process exitcode."""
        instance.exitcode

    def test_get_main_proc(self, instance):
        r"""Test get_main_proc."""
        instance.get_main_proc()

    def test_kill(self, instance, task_method):
        r"""Test kill."""
        if task_method == 'process':
            instance.start()
        instance.kill()
        instance.exitcode

    def test_flag_manipulation(self, instance):
        r"""Test flag manipulation."""
        instance.set_flag_attr('error_flag')
        assert(instance.check_flag_attr('error_flag'))
        instance.clear_flag_attr('error_flag')
        assert(not instance.check_flag_attr('error_flag'))


class TestYggProcessFork(TestYggTask):
    r"""Test basic behavior of YggTask for forked process."""

    @pytest.fixture(scope="class", autouse=True, params=['process'])
    def task_method(self, request):
        r"""Method that should be used for the generatd task."""
        return request.param
    
    @pytest.fixture(scope="class")
    def instance_kwargs(self, timeout, polling_interval, task_method):
        r"""Keyword arguments for a new instance of the tested class."""
        return dict(timeout=timeout, sleeptime=polling_interval,
                    target=multitasking.test_target_sleep,
                    task_method=task_method,
                    context=multitasking.mp_ctx)



import pytest
from tests.metaschema.datatypes.test_MetaschemaType import (
    TestMetaschemaType as base_class)


class TestAnyMetaschemaType(base_class):
    r"""Test class for AnyMetaschemaType class."""

    _mod = 'yggdrasil.metaschema.datatypes.AnyMetaschemaType'
    _cls = 'AnyMetaschemaType'

    @pytest.fixture(scope="class")
    def value(self):
        r"""dict: Test value."""
        return {'a': int(1), 'b': float(1)}

    @pytest.fixture(scope="class")
    def valid_encoded(self, typedef_base, value, python_class):
        r"""list: Encoded objects that are valid under this type."""
        return [dict(typedef_base,
                     type=python_class.name,
                     temptype={'type': 'int'})]
    
    @pytest.fixture(scope="class")
    def valid_decoded(self, value):
        r"""list: Objects that are valid under this type."""
        return [value, object]
        
    @pytest.fixture(scope="class")
    def invalid_validate(self):
        r"""list: Objects that are invalid under this type."""
        return []
    
    @pytest.fixture(scope="class")
    def invalid_encoded(self):
        r"""list: Encoded objects that are invalid under this type."""
        return []

    @pytest.fixture(scope="class")
    def compatible_objects(self, value):
        r"""list: Objects that are compatible with this type."""
        return [(value, value, None)]
    
    def test_decode_data_errors(self, python_class):
        r"""Test errors in decode_data."""
        with pytest.raises(ValueError):
            python_class.decode_data('hello', None)



import pytest
import copy
from yggdrasil import tools, platform, constants
from yggdrasil.components import import_component
from yggdrasil.drivers.ConnectionDriver import ConnectionDriver
from yggdrasil.communication import new_comm, CommBase
from tests import timeout_decorator
from tests.drivers.test_Driver import TestDriver as base_class


_default_comm = tools.get_default_comm()


@pytest.mark.skipif(platform._is_win,
                    reason=("Temp skip connection tests on windows for "
                            "time's sake."))
@pytest.mark.suite("connections")
class TestConnectionDriver(base_class):
    r"""Test class for the ConnectionDriver class."""

    _component_type = 'connection'
    parametrize_commtype = [_default_comm]

    @pytest.fixture(scope="class")
    def component_subtype(self):
        r"""Subtype of component being tested."""
        return 'connection'

    @pytest.fixture(scope="class")
    def default_comm(self):
        r"""str: Name of the default communicator."""
        return _default_comm

    @pytest.fixture(scope="class")
    def commtype(self, request):
        r"""str: Name of the communicator being tested."""
        return request.param

    @pytest.fixture(scope="class")
    def icomm_name(self, default_comm):
        r"""str: Name of the input communicator being tested."""
        return default_comm
    
    @pytest.fixture(scope="class")
    def ocomm_name(self, default_comm):
        r"""str: Name of the output communicator being tested."""
        return default_comm

    @pytest.fixture(scope="class", autouse=True)
    def running_service_for_comms(self, icomm_name, ocomm_name,
                                  running_service):
        if 'rest' in [icomm_name, ocomm_name]:
            with running_service('flask') as cli:
                yield cli
        else:
            yield None
            
    @pytest.fixture
    def maxMsgSize(self, instance):
        r"""int: Maximum message size."""
        return min(instance.icomm.maxMsgSize,
                   instance.ocomm.maxMsgSize)

    @pytest.fixture(scope="class")
    def is_output(self, default_comm, ocomm_name):
        r"""bool: True if the connection is for output."""
        return (ocomm_name != default_comm)

    @pytest.fixture(scope="class")
    def test_msg(self, testing_options):
        r"""str: Test message that should be used for any send/recv tests."""
        return testing_options['msg']
        
    @pytest.fixture(scope="class")
    def nested_result(self, nested_approx):
        r"""Convert a sent object into a received one."""
        def nested_result_w(obj):
            return nested_approx(obj)
        return nested_result_w

    @pytest.fixture(scope="class")
    def testing_options(self, python_class, options, is_output,
                        icomm_python_class, ocomm_python_class):
        r"""Testing options."""
        if 'explicit_testing_options' in options:
            return copy.deepcopy(options['explicit_testing_options'])
        if is_output:
            out = ocomm_python_class.get_testing_options(**options)
        else:
            out = icomm_python_class.get_testing_options(**options)
        return out
    
    @pytest.fixture
    def inputs(self, icomm_name, testing_options, is_output):
        r"""list: List of keyword arguments for connection input comms."""
        out = [{'name': icomm_name, 'commtype': icomm_name}]
        if not is_output:
            out[0].update(testing_options['kwargs'])
        return out

    @pytest.fixture
    def outputs(self, ocomm_name, testing_options, is_output):
        r"""list: List of keyword arguments for connection output comms."""
        out = [{'name': ocomm_name, 'commtype': ocomm_name}]
        if is_output:
            out[0].update(testing_options['kwargs'])
        return out

    @pytest.fixture(scope="class")
    def icomm_python_class(self, icomm_name):
        r"""class: Class used for connection input comm."""
        return import_component('comm', icomm_name)

    @pytest.fixture(scope="class")
    def ocomm_python_class(self, ocomm_name):
        r"""class: Class used for connection output comm."""
        return import_component('comm', ocomm_name)

    @pytest.fixture
    def send_comm_kwargs(self, instance, icomm_name):
        r"""dict: Keyword arguments for send comm."""
        out = instance.icomm.opp_comm_kwargs()
        if icomm_name == 'value':
            out['direction'] = 'recv'
        return out

    @pytest.fixture
    def recv_comm_kwargs(self, instance):
        r"""dict: Keyword arguments for recv comm."""
        return instance.ocomm.opp_comm_kwargs()

    @pytest.fixture
    def instance_kwargs(self, testing_options, timeout, working_dir,
                        polling_interval, namespace, inputs, outputs,
                        icomm_name, icomm_python_class,
                        ocomm_name, ocomm_python_class):
        r"""Keyword arguments for a new instance of the tested class."""
        if not (ocomm_python_class.is_installed(language='python')
                and icomm_python_class.is_installed(language='python')):
            pytest.skip(
                f"one of the tested comms ({icomm_name}="
                f"{icomm_python_class.is_installed(language='python')}, "
                f"{ocomm_name}="
                f"{ocomm_python_class.is_installed(language='python')}) "
                f"is not installed")
        return dict(testing_options.get('kwargs', {}),
                    yml={'working_dir': working_dir},
                    timeout=timeout, sleeptime=polling_interval,
                    namespace=namespace, inputs=inputs, outputs=outputs)

    @pytest.fixture
    def msg_long(self, test_msg, maxMsgSize):
        r"""str: Small test message for sending."""
        if isinstance(test_msg, bytes):
            out = test_msg + (maxMsgSize * b'0')
        else:  # pragma: debug
            out = test_msg
        return out

    @pytest.fixture
    def send_comm(self, name, send_comm_kwargs, close_comm):
        r"""CommBase: communicator for sending messages to the driver."""
        out = new_comm(name, **send_comm_kwargs)
        try:
            yield out
        finally:
            close_comm(out)
    
    @pytest.fixture
    def recv_comm(self, name, recv_comm_kwargs, close_comm):
        r"""CommBase: communicator for receiving messages from the driver."""
        out = new_comm(name, **recv_comm_kwargs)
        try:
            yield out
        finally:
            close_comm(out)
    
    def test_manipulate_shared_event(self, instance):
        r"""Test setting and clearing events in the shared dictionary."""
        instance.set_flag_attr('_skip_after_loop')
        assert(instance.check_flag_attr('_skip_after_loop'))
        instance.clear_flag_attr('_skip_after_loop')
        assert(not instance.check_flag_attr('_skip_after_loop'))

    def test_send_recv_closed(self, instance, send_comm, recv_comm, test_msg):
        r"""Test sending/receiving with queues closed."""
        instance.close_comm()
        send_comm.close()
        recv_comm.close()
        assert(instance.is_comm_closed)
        assert(send_comm.is_closed)
        assert(recv_comm.is_closed)
        flag = instance.send_message(CommBase.CommMessage(args=test_msg))
        assert(not flag)
        flag = instance.recv_message()
        if instance.icomm._commtype != 'value':
            assert(not flag)
        # Short
        if instance.icomm._commtype != 'value':
            flag = send_comm.send(test_msg)
            assert(not flag)
        flag, ret = recv_comm.recv()
        if instance.icomm._commtype != 'value':
            assert(not flag)
            assert(ret is None)
        # Long
        if instance.icomm._commtype != 'value':
            flag = send_comm.send_nolimit(test_msg)
            assert(not flag)
        flag, ret = recv_comm.recv_nolimit()
        if instance.icomm._commtype != 'value':
            assert(not flag)
            assert(ret is None)
        instance.confirm_output(timeout=1.0)

    def test_error_init_ocomm(self, monkeypatch, python_class,
                              instance_args, instance_kwargs,
                              MagicTestError):
        r"""Test forwarding of error from init of ocomm."""
        orig = python_class._init_single_comm

        def error_init_single_comm(self, direction, *args):
            if direction == 'output':
                x = MagicTestError()
                x.instance = self
                raise x
            else:
                return orig(self, direction, *args)
        with monkeypatch.context() as m:
            m.setattr(python_class, '_init_single_comm',
                      error_init_single_comm)
            with pytest.raises(MagicTestError) as excinfo:
                python_class(*instance_args, **instance_kwargs)
            excinfo.value.instance.disconnect()

    def test_error_open_icomm(self, monkeypatch, instance,
                              MagicTestError, magic_error_replacement):
        r"""Test fowarding of error from open of icomm."""
        with monkeypatch.context() as m:
            if hasattr(instance.icomm, '_wrapped'):
                m.setattr(instance.icomm._wrapped, 'open',
                          magic_error_replacement)
            else:
                m.setattr(instance.icomm, 'open', magic_error_replacement)
            with pytest.raises(MagicTestError):
                instance.open_comm()
            assert(instance.icomm.is_closed)

    def test_error_close_icomm(self, monkeypatch, instance,
                               MagicTestError, magic_error_replacement):
        r"""Test forwarding of error from close of icomm."""
        instance.open_comm()
        with monkeypatch.context() as m:
            if hasattr(instance.icomm, '_wrapped'):
                m.setattr(instance.icomm._wrapped, '_close',
                          magic_error_replacement)
            else:
                m.setattr(instance.icomm, 'close',
                          magic_error_replacement)
            with pytest.raises(MagicTestError):
                instance.close_comm()
        assert(instance.ocomm.is_closed)
        instance.icomm.close()
        assert(instance.icomm.is_closed)
        
    def test_error_close_ocomm(self, monkeypatch, instance,
                               MagicTestError, magic_error_replacement):
        r"""Test forwarding of error from close of ocomm."""
        instance.open_comm()
        with monkeypatch.context() as m:
            m.setattr(instance.ocomm, '_close',
                      magic_error_replacement)
            with pytest.raises(MagicTestError):
                instance.close_comm()
        assert(instance.icomm.is_closed)
        instance.ocomm.close()
        assert(instance.ocomm.is_closed)

    def test_error_open_fails(self, monkeypatch, instance):
        r"""Test error raised when comms fail to open."""
        def empty_open(*args, **kwargs):
            pass
        with monkeypatch.context() as m:
            def replace_open(x):
                if hasattr(x, '_wrapped'):
                    replace_open(x._wrapped)
                elif x._commtype.startswith('rmq'):
                    x.close()
                else:
                    m.setattr(x, 'open', empty_open)
            replace_open(instance.icomm)
            replace_open(instance.ocomm)
            m.setattr(instance, 'timeout', instance.sleeptime / 2.0)
            with pytest.raises(Exception):
                instance.start()
        instance.close_comm()
        assert(instance.is_comm_closed)

    @pytest.fixture
    def nmsg_recv(self, icomm_name, testing_options):
        r"""Expected number of messages."""
        if icomm_name == 'value':
            return testing_options['kwargs']['count']
        return 1

    @pytest.fixture
    def before_instance_started(self, send_comm, recv_comm):
        r"""Actions performed after teh instance is created, but before it
        is started."""
        def before_instance_started_w(x):
            pass
        return before_instance_started_w
    
    @pytest.fixture
    def after_instance_started(self, wait_on_function, recv_comm):
        r"""Action taken after the instance is started, but before tests
        begin."""
        def after_instance_started_w(instance):
            wait_on_function(lambda: instance.is_valid)
            recv_comm.drain_server_signon_messages()
        return after_instance_started_w
        
    def test_init_del(self, started_instance):
        r"""Test driver creation and deletion."""
        started_instance.printStatus(verbose=True)
        started_instance.printStatus(verbose=True, return_str=True)

    def test_early_close(self, started_instance):
        r"""Test early deletion of message queue."""
        started_instance.close_comm()
        started_instance.open_comm()
        assert(started_instance.is_comm_closed)

    @timeout_decorator(timeout=600)
    def test_send_recv(self, started_instance, send_comm, recv_comm,
                       test_msg, nmsg_recv, timeout, icomm_name,
                       nested_result):
        r"""Test sending/receiving small message."""
        try:
            if started_instance.icomm._commtype != 'value':
                flag = send_comm.send(test_msg)
                assert(flag)
            # started_instance.sleep()
            # assert(recv_comm.n_msg == 1)
            for i in range(nmsg_recv):
                flag, msg_recv = recv_comm.recv(timeout=timeout)
                assert(flag)
                assert(msg_recv == nested_result(test_msg))
            if icomm_name != 'value':
                assert(started_instance.n_msg == 0)
        except BaseException:  # pragma: debug
            send_comm.printStatus()
            started_instance.printStatus(verbose=True)
            recv_comm.printStatus()
            raise

    @timeout_decorator(timeout=600)
    def test_send_recv_nolimit(self, started_instance, send_comm, recv_comm,
                               msg_long, maxMsgSize, nmsg_recv, timeout,
                               nested_result):
        r"""Test sending/receiving large message."""
        try:
            if started_instance.icomm._commtype != 'value':
                assert(len(msg_long) > maxMsgSize)
                flag = send_comm.send_nolimit(msg_long)
                assert(flag)
            for i in range(nmsg_recv):
                flag, msg_recv = recv_comm.recv_nolimit(timeout=timeout)
                assert(flag)
                assert(msg_recv == nested_result(msg_long))
        except BaseException:  # pragma: debug
            send_comm.printStatus()
            started_instance.printStatus(verbose=True)
            recv_comm.printStatus()
            raise

    @pytest.fixture
    def assert_before_stop(self, icomm_name, instance):
        r"""Assertions to make before stopping the driver instance."""
        def assert_before_stop_w():
            if icomm_name != 'value':
                assert(instance.is_comm_open)
        return assert_before_stop_w

    # TODO: This fails with ZMQ
    # @pytest.fixture
    # def run_before_terminate(self, send_comm, test_msg):
    #     r"""Commands to run while the instance is running, before terminate."""
    #     def run_before_terminate_w():
    #         send_comm.send(test_msg)
    #     return run_before_terminate_w

    @pytest.fixture
    def assert_after_terminate(self, started_instance):
        r"""Assertions to make after terminating the driver instance."""
        def assert_after_terminate_w():
            assert(not started_instance.is_alive())
            assert(started_instance.is_comm_closed)
        return assert_after_terminate_w


class TestConnectionDriverFork(TestConnectionDriver):
    r"""Test class for the ConnectionDriver class between fork comms."""

    @pytest.fixture(scope="class")
    def ncomm_input(self):
        r"""int: Number of input comms."""
        return 2

    @pytest.fixture(scope="class")
    def ncomm_output(self):
        r"""int: Number of output comms."""
        return 1

    @pytest.fixture(scope="class")
    def nmsg_recv(self, ncomm_input, ncomm_output):
        r"""int: Number of messages expected."""
        return ncomm_input * ncomm_output

    @pytest.fixture
    def inputs(self, ncomm_input):
        r"""list: List of keyword arguments for connection input comms."""
        return [None for _ in range(ncomm_input)]


invalid_translate = True


class TestConnectionDriverTranslate(TestConnectionDriver):
    r"""Test class for the ConnectionDriver class with translator."""

    test_send_recv_nolimit = None

    @pytest.fixture
    def instance_kwargs(self, testing_options, timeout, working_dir,
                        polling_interval, namespace, inputs, outputs):
        r"""Keyword arguments for a new instance of the tested class."""
        return dict(testing_options.get('kwargs', {}),
                    yml={'working_dir': working_dir},
                    timeout=timeout, sleeptime=polling_interval,
                    namespace=namespace, inputs=inputs, outputs=outputs,
                    translator={'transformtype': 'select_fields',
                                'selected': ['a'],
                                'single_as_scalar': True},
                    onexit='printStatus')

    @pytest.fixture(scope="class")
    def test_msg(self):
        r"""str: Test message that should be used for any send/recv tests."""
        return {'a': int(1), 'b': float(2)}

    @pytest.fixture(scope="class")
    def nested_result(self, nested_approx):
        r"""Convert a sent object into a received one."""
        def nested_result_w(obj):
            return nested_approx(obj['a'])
        return nested_result_w
    

class TestConnectionDriverIterate(TestConnectionDriver):
    r"""Test class for the ConnectionDriver class with iteration."""

    test_send_recv_nolimit = None

    @pytest.fixture
    def instance_kwargs(self, testing_options, timeout, working_dir,
                        polling_interval, namespace, inputs, outputs):
        r"""Keyword arguments for a new instance of the tested class."""
        return dict(testing_options.get('kwargs', {}),
                    yml={'working_dir': working_dir},
                    timeout=timeout, sleeptime=polling_interval,
                    namespace=namespace, inputs=inputs, outputs=outputs,
                    translator=[{'transformtype': 'select_fields',
                                 'selected': ['a', 'b']},
                                {'transformtype': 'iterate'}],
                    onexit='printStatus')

    @pytest.fixture(scope="class")
    def test_msg(self):
        r"""str: Test message that should be used for any send/recv tests."""
        return {'a': int(1), 'b': 'hello', 'c': float(2)}

    @timeout_decorator(timeout=600)
    def test_send_recv(self, started_instance, send_comm, recv_comm, test_msg,
                       timeout, nested_result):
        r"""Test sending/receiving small message."""
        msg = test_msg
        flag = send_comm.send(msg)
        assert(flag)
        for imsg in [v for k, v in msg.items() if k in ['a', 'b']]:
            flag, msg_recv = recv_comm.recv(timeout)
            assert(flag)
            assert(msg_recv == nested_result(imsg))
        assert(started_instance.n_msg == 0)


class TestConnectionDriverProcess(TestConnectionDriver):
    r"""Test class for the TestConnectionDriver using process."""

    test_send_recv_closed = None

    @pytest.fixture(scope="class")
    def icomm_name(self):
        r"""str: Name of the input communicator being tested."""
        return 'buffer'
    
    @pytest.fixture
    def instance_kwargs(self, testing_options, timeout, working_dir,
                        polling_interval, namespace, inputs, outputs):
        r"""Keyword arguments for a new instance of the tested class."""
        return dict(testing_options.get('kwargs', {}),
                    yml={'working_dir': working_dir},
                    timeout=timeout, sleeptime=polling_interval,
                    namespace=namespace, inputs=inputs, outputs=outputs,
                    task_method='process')


def test_ConnectionDriverOnexit_errors():
    r"""Test that errors are raised for invalid onexit."""
    with pytest.raises(ValueError):
        ConnectionDriver('test', onexit='invalid')


def test_ConnectionDriverTranslate_errors():
    r"""Test that errors are raised for invalid translators."""
    assert(not hasattr(invalid_translate, '__call__'))
    with pytest.raises(ValueError):
        ConnectionDriver('test', translator=invalid_translate)


_comm_types = sorted(
    [x for x in constants.COMPONENT_REGISTRY['comm']['subtypes'].keys()
     if x not in [_default_comm, 'mpi']])


class TestOutputDriver(TestConnectionDriver):
    r"""Test output drivers for supported comm types."""

    parametrize_commtype = [x for x in _comm_types if x not in ['value']]

    @pytest.fixture(scope="class")
    def component_subtype(self):
        r"""Subtype of component being tested."""
        return 'output'
    
    @pytest.fixture
    def instance_args(self, name):
        r"""Arguments for a new instance of the tested class."""
        return (name, 'test')
    
    @pytest.fixture(scope="class")
    def ocomm_name(self, commtype):
        r"""str: Name of the output communicator being tested."""
        return commtype


class TestInputDriver(TestConnectionDriver):
    r"""Test input drivers for supported comm types."""

    parametrize_commtype = _comm_types

    @pytest.fixture(scope="class")
    def component_subtype(self):
        r"""Subtype of component being tested."""
        return 'input'
    
    @pytest.fixture
    def instance_args(self, name):
        r"""Arguments for a new instance of the tested class."""
        return (name, 'test')
    
    @pytest.fixture(scope="class")
    def icomm_name(self, commtype):
        r"""str: Name of the input communicator being tested."""
        return commtype



import pytest
from tests.metaschema.properties.test_MetaschemaProperty import (
    TestMetaschemaProperty as base_class)


class TestPropertiesMetaschemaProperty(base_class):
    r"""Test class for PropertiesMetaschemaProperty class."""
    
    _mod = 'yggdrasil.metaschema.properties.JSONObjectMetaschemaProperties'
    _cls = 'PropertiesMetaschemaProperty'

    @pytest.fixture(scope="class")
    def ele(self):
        r"""Test element"""
        return 'abc'

    @pytest.fixture(scope="class")
    def valid_value(self, ele):
        r"""Valid test value."""
        return {x: int(i) for i, x in enumerate(ele)}

    @pytest.fixture(scope="class")
    def valid_base(self, ele):
        r"""Base class for valid arrays."""
        return {x: {'type': 'int'} for x in ele}

    @pytest.fixture(scope="class")
    def invalid_type(self, ele):
        r"""Invalid type."""
        return {x: {'type': 'float'} for x in ele}

    @pytest.fixture(scope="class")
    def invalid_keys(self, ele):
        r"""Invalid keys."""
        return {x: {'type': 'int'} for x in ele[:-1]}

    @pytest.fixture(scope="class")
    def valid(self, valid_value, valid_base):
        r"""Objects that are valid."""
        return [(valid_value, valid_base)]

    @pytest.fixture(scope="class")
    def invalid(self, ele, valid_base):
        r"""Objects that are invalid."""
        return [({x: float(i) for i, x in enumerate(ele)}, valid_base)]

    @pytest.fixture(scope="class")
    def valid_compare(self, valid_base, invalid_keys):
        r"""Objects that successfully compare."""
        return [(valid_base, valid_base),
                (valid_base, invalid_keys)]

    @pytest.fixture(scope="class")
    def invalid_compare(self, valid_base, invalid_type, invalid_keys):
        r"""Objects that do not successfully compare."""
        return [(invalid_type, valid_base),
                (invalid_keys, valid_base)]
